import torch

import numpy as np
import torch.nn as nn
import torch.distributed as dist
import torch.nn.functional as F

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Optional
from torch.optim import Optimizer
from cftool.misc import update_dict
from cftool.misc import shallow_copy_dict
from cftool.array import to_device
from cftool.array import l2_normalize
from cftool.types import tensor_dict_type
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.cuda.amp import GradScaler

from ..encoder import run_encoder
from ..encoder import Encoder1DMixin
from ....data import CVLoader
from ....protocol import ITrainer
from ....protocol import StepOutputs
from ....protocol import TrainerState
from ....protocol import MetricsOutputs
from ....constants import LOSS_KEY
from ....constants import INPUT_KEY
from ....constants import LATENT_KEY
from ....misc.toolkit import get_world_size
from ....misc.toolkit import has_batch_norms
from ....misc.internal_.register import register_custom_module
from ....misc.internal_.register import CustomModule


def _get_dino_defaults(name: str) -> Dict[str, Any]:
    if name == "vit":
        return {"patch_size": 16, "drop_path_rate": 0.1}
    return {}


class Scheduler:
    def __init__(self, values: np.ndarray):
        self.values = values
        self.max_idx = len(values) - 1

    def __getitem__(self, index: int) -> Any:
        return self.values[min(index, self.max_idx)]


def cosine_scheduler(
    base_value: float,
    final_value: float,
    epochs: int,
    num_step_per_epoch: int,
    warmup_epochs: int = 0,
    start_warmup_value: int = 0,
) -> Scheduler:
    warmup_schedule = np.array([])
    warmup_iters = warmup_epochs * num_step_per_epoch
    if warmup_epochs > 0:
        warmup_schedule = np.linspace(start_warmup_value, base_value, warmup_iters)
    iters = np.arange(epochs * num_step_per_epoch - warmup_iters)
    diff = base_value - final_value
    schedule = final_value + 0.5 * diff * (1.0 + np.cos(np.pi * iters / len(iters)))
    schedule = np.concatenate((warmup_schedule, schedule))
    assert len(schedule) == epochs * num_step_per_epoch
    return Scheduler(schedule)


class MultiCropWrapper(nn.Module):
    def __init__(self, backbone: Encoder1DMixin, head: nn.Module):
        super().__init__()
        backbone.fc, backbone.head = nn.Identity(), nn.Identity()
        self.backbone = backbone
        self.head = head

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        *,
        img_end_idx: Optional[int] = None,
        **kwargs: Any,
    ) -> Tensor:
        img_crops = batch[INPUT_KEY]
        if not isinstance(img_crops, list):
            img_crops = batch[INPUT_KEY] = [img_crops]
        if img_end_idx is not None:
            img_crops = img_crops[:img_end_idx]
        idx_crops = torch.cumsum(
            torch.unique_consecutive(
                torch.tensor([img_crop.shape[-1] for img_crop in img_crops]),
                return_counts=True,
            )[1],
            0,
        )
        outputs = []
        start_idx = 0
        for end_idx in idx_crops:
            local_batch = shallow_copy_dict(batch)
            local_batch[INPUT_KEY] = torch.cat(img_crops[start_idx:end_idx])
            idx_rs = run_encoder(self.backbone, batch_idx, local_batch, state, **kwargs)
            idx_out = idx_rs[LATENT_KEY]
            if isinstance(idx_out, tuple):
                idx_out = idx_out[0]
            outputs.append(idx_out)
            start_idx = end_idx
        return self.head(torch.cat(outputs))


class DINOHead(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        batch_norm: bool = False,
        norm_last_layer: bool = True,
        *,
        num_layers: int = 3,
        latent_dim: int = 2048,
        bottleneck_dim: int = 256,
    ):
        super().__init__()
        num_layers = max(num_layers, 1)
        if num_layers == 1:
            self.mapping = nn.Linear(in_dim, bottleneck_dim)
        else:
            blocks = [nn.Linear(in_dim, latent_dim)]
            if batch_norm:
                blocks.append(nn.BatchNorm1d(latent_dim))
            blocks.append(nn.GELU())
            for _ in range(num_layers - 2):
                blocks.append(nn.Linear(latent_dim, latent_dim))
                if batch_norm:
                    blocks.append(nn.BatchNorm1d(latent_dim))
                blocks.append(nn.GELU())
            blocks.append(nn.Linear(latent_dim, bottleneck_dim))
            self.mapping = nn.Sequential(*blocks)
        self.apply(self._init_weights)
        last = nn.Linear(bottleneck_dim, out_dim, bias=False)
        self.last_layer = nn.utils.weight_norm(last)
        self.last_layer.weight_g.data.fill_(1)
        if norm_last_layer:
            self.last_layer.weight_g.requires_grad = False

    def _init_weights(self, m: nn.Module) -> None:
        if isinstance(m, nn.Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, net: Tensor) -> Tensor:
        net = self.mapping(net)
        net = nn.functional.normalize(net, dim=-1, p=2)
        net = self.last_layer(net)
        return net


class DINOLoss(nn.Module):
    center: torch.Tensor

    def __init__(
        self,
        out_dim: int,
        teacher_temp: float,
        warmup_teacher_temp: float,
        warmup_teacher_temp_epochs: int,
        teacher_temp_epochs: int,
        *,
        student_temp: float = 0.1,
        center_momentum: float = 0.9,
    ):
        super().__init__()
        self.student_temp = student_temp
        self.center_momentum = center_momentum
        self.register_buffer("center", torch.zeros(1, out_dim))
        teacher_temp_constant_epochs = teacher_temp_epochs - warmup_teacher_temp_epochs
        self.teacher_temp_schedule = Scheduler(
            np.concatenate(
                (
                    np.linspace(
                        warmup_teacher_temp,
                        teacher_temp,
                        warmup_teacher_temp_epochs,
                    ),
                    np.ones(teacher_temp_constant_epochs) * teacher_temp,
                )
            )
        )
        self.num_epochs = teacher_temp_epochs

    def forward(
        self,
        epoch: int,
        num_crops: int,
        student_output: Tensor,
        teacher_output: Tensor,
    ) -> Tensor:
        student_logits = student_output / self.student_temp
        student_logits_list = student_logits.chunk(num_crops)

        temp = self.teacher_temp_schedule[epoch]
        teacher_logits = F.softmax((teacher_output - self.center) / temp, dim=-1)
        teacher_logits_list = teacher_logits.detach().chunk(2)

        total_loss = 0.0
        num_loss_terms = 0
        for it, t_logit in enumerate(teacher_logits_list):
            for iv, v_logit in enumerate(student_logits_list):
                if iv == it:
                    continue
                loss = torch.sum(-t_logit * F.log_softmax(v_logit, dim=-1), dim=-1)
                total_loss += loss.mean()
                num_loss_terms += 1
        total_loss /= num_loss_terms
        self.update_center(teacher_output)
        return total_loss

    @torch.no_grad()
    def update_center(self, teacher_output: Tensor) -> None:
        batch_center = torch.sum(teacher_output, dim=0, keepdim=True)
        if dist.is_initialized():
            dist.all_reduce(batch_center)
        batch_center = batch_center / (len(teacher_output) * get_world_size())
        m = self.center_momentum
        self.center = self.center * m + batch_center * (1.0 - m)


class DINOEvaluateLoss:
    def __init__(self, train_loss: DINOLoss):
        self.train_loss = train_loss

    def __call__(
        self,
        epoch: int,
        student_output: Tensor,
        teacher_output: Tensor,
    ) -> float:
        s_logits = student_output / self.train_loss.student_temp
        temp = self.train_loss.teacher_temp_schedule[epoch]
        centered = teacher_output - self.train_loss.center
        t_logits = F.softmax(centered / temp, dim=-1)
        loss = torch.sum(-t_logits * F.log_softmax(s_logits, dim=-1), dim=-1).mean()
        return loss.item()


@register_custom_module(
    "dino",
    custom_params_groups=True,
    custom_ddp_initialization=True,
)
class DINO(CustomModule):
    lr_schedule: Optional[Scheduler]
    wd_schedule: Optional[Scheduler]
    momentum_schedule: Optional[Scheduler]

    def __init__(
        self,
        encoder1d: str = "vit",
        encoder1d_config: Optional[Dict[str, Any]] = None,
        student_specific: Optional[Dict[str, Any]] = None,
        teacher_specific: Optional[Dict[str, Any]] = None,
        *,
        out_dim: int = 65536,
        use_bn_in_head: bool = False,
        norm_last_layer: bool = True,
        teacher_temp: float = 0.07,
        momentum_teacher: float = 0.996,
        warmup_teacher_temp: float = 0.04,
        warmup_teacher_temp_epochs: int = 30,
        teacher_temp_epochs: int,
        freeze_last_layer: int = 1,
        weight_decay: float = 0.04,
        weight_decay_end: float = 0.4,
        warmup_epochs: int = 10,
    ):
        super().__init__()
        base = update_dict(encoder1d_config or {}, _get_dino_defaults(encoder1d))
        student_cfg = update_dict(student_specific or {}, shallow_copy_dict(base))
        teacher_cfg = update_dict(teacher_specific or {}, shallow_copy_dict(base))
        student = Encoder1DMixin.make(encoder1d, student_cfg)
        teacher = Encoder1DMixin.make(encoder1d, teacher_cfg)
        self.ddp_student = self.ddp_teacher = None
        self.student = MultiCropWrapper(
            student,
            DINOHead(
                student.latent_dim,
                out_dim,
                use_bn_in_head,
                norm_last_layer,
            ),
        )
        self.teacher = MultiCropWrapper(
            teacher,
            DINOHead(teacher.latent_dim, out_dim, use_bn_in_head),
        )
        self.freeze_last_layer = freeze_last_layer
        self.teacher.load_state_dict(self.student.state_dict())
        self.loss = DINOLoss(
            out_dim,
            teacher_temp,
            warmup_teacher_temp,
            warmup_teacher_temp_epochs,
            teacher_temp_epochs,
        )
        self.evaluate_loss = DINOEvaluateLoss(self.loss)
        self.momentum_teacher = momentum_teacher
        self.teacher_temp_epochs = teacher_temp_epochs
        self.weight_decay = weight_decay
        self.weight_decay_end = weight_decay_end
        self.warmup_epochs = warmup_epochs
        self.lr_schedule = None
        self.wd_schedule = None
        self.momentum_schedule = None

    @property
    def student_for_training(self) -> MultiCropWrapper:
        return self.ddp_student or self.student

    @property
    def teacher_for_training(self) -> MultiCropWrapper:
        return self.ddp_teacher or self.teacher

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        res = run_encoder(self.student.backbone, batch_idx, batch, state, **kwargs)
        net = res[LATENT_KEY]
        net = l2_normalize(net)
        return {LATENT_KEY: net}

    def onnx_forward(self, batch: tensor_dict_type) -> Any:
        inp = batch[INPUT_KEY]
        net = self.get_latent(inp, determinate=True)
        return net.view(inp.shape[0], self.student.backbone.latent_dim)

    def summary_forward(self, batch_idx: int, batch: tensor_dict_type) -> None:
        self.student(batch_idx, to_device(batch, self.device))

    def get_latent(self, net: Tensor, **kwargs: Any) -> Tensor:
        return self.forward(0, {INPUT_KEY: net}, **kwargs)[LATENT_KEY]

    def get_logits(self, net: Tensor) -> Tensor:
        return self.student(0, {INPUT_KEY: net})

    def state_dict(
        self,
        destination: Any = None,
        prefix: str = "",
        keep_vars: bool = False,
    ) -> Any:
        states = super().state_dict(destination, prefix, keep_vars)
        for k in list(states.keys()):
            if k.startswith("ddp"):
                states.pop(k)
        return states

    def _get_outputs(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: TrainerState,
        forward_kwargs: Dict[str, Any],
    ) -> tensor_dict_type:
        teacher_output = self.teacher_for_training(
            batch_idx,
            batch,
            state,
            img_end_idx=2,
            **forward_kwargs,
        )
        student_output = self.student_for_training(
            batch_idx,
            batch,
            state,
            **forward_kwargs,
        )
        return {"student": student_output, "teacher": teacher_output}

    def _get_loss(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: TrainerState,
        use_amp: bool,
        forward_kwargs: Dict[str, Any],
    ) -> Tuple[tensor_dict_type, Tensor]:
        with torch.cuda.amp.autocast(enabled=use_amp):
            outputs = self._get_outputs(batch_idx, batch, state, forward_kwargs)
            epoch = state.epoch
            num_crops = len(batch[INPUT_KEY])
            student_output = outputs["student"]
            teacher_output = outputs["teacher"]
            loss = self.loss(epoch, num_crops, student_output, teacher_output)
        return outputs, loss

    def train_step(  # type: ignore
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: TrainerState,
        optimizers: Dict[str, Optimizer],
        use_amp: bool,
        grad_scaler: GradScaler,
        trainer: ITrainer,
        forward_kwargs: Dict[str, Any],
    ) -> StepOutputs:
        if self.lr_schedule is None:
            self.lr_schedule = cosine_scheduler(
                self.lr * (len(batch[INPUT_KEY][0]) * get_world_size()) / 256.0,  # type: ignore
                self.min_lr,
                self.teacher_temp_epochs,
                state.num_step_per_epoch,
                warmup_epochs=self.warmup_epochs,
            )
        if self.wd_schedule is None:
            self.wd_schedule = cosine_scheduler(
                self.weight_decay,
                self.weight_decay_end,
                self.teacher_temp_epochs,
                state.num_step_per_epoch,
            )
        # manual scheduling
        optimizer = optimizers["all"]
        for i, param_group in enumerate(optimizer.param_groups):
            param_group["lr"] = self.lr_schedule[state.step]
            if i == 0:
                param_group["weight_decay"] = self.wd_schedule[state.step]
        # forward pass
        rs, loss = self._get_loss(batch_idx, batch, state, use_amp, forward_kwargs)
        # backward pass
        optimizer.zero_grad()
        grad_scaler.scale(loss).backward()
        # clip norm
        if trainer.clip_norm > 0.0:
            grad_scaler.unscale_(optimizer)
            nn.utils.clip_grad_norm_(
                self.student_for_training.parameters(),
                max_norm=trainer.clip_norm,
            )
        # freeze last layer
        if state.epoch <= self.freeze_last_layer:
            for n, p in self.student.named_parameters():
                if "last_layer" in n:
                    p.grad = None
        # update parameters
        grad_scaler.step(optimizer)
        grad_scaler.update()
        # update momentum teacher
        if self.momentum_schedule is None:
            self.momentum_schedule = cosine_scheduler(
                self.momentum_teacher,
                1.0,
                self.teacher_temp_epochs,
                state.num_step_per_epoch,
            )
        with torch.no_grad():
            m = self.momentum_schedule[state.step]
            for param_q, param_k in zip(
                self.student.parameters(),
                self.teacher.parameters(),
            ):
                param_k.data.mul_(m).add_((1.0 - m) * param_q.detach().data)
        # return
        return StepOutputs(rs, {LOSS_KEY: loss.item()})

    def evaluate_step(  # type: ignore
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: TrainerState,
    ) -> MetricsOutputs:
        outputs = self._get_outputs(batch_idx, batch, state, {})
        loss = self.evaluate_loss(
            state.epoch,
            outputs["student"],
            outputs["teacher"],
        )
        return MetricsOutputs(
            -loss,
            {
                "loss": loss,
                "lr": self.lr_schedule[state.step],  # type: ignore
                "wd": self.wd_schedule[state.step],  # type: ignore
            },
        )

    def params_groups(self, m: nn.Module) -> Any:
        regularized = []
        bias_and_norm = []
        for name, param in m.named_parameters():
            if not param.requires_grad:
                continue
            if name.endswith(".bias") or len(param.shape) == 1:
                bias_and_norm.append(param)
            else:
                regularized.append(param)
        return [{"params": regularized}, {"params": bias_and_norm, "weight_decay": 0.0}]

    def init_ddp(self) -> None:
        if has_batch_norms(self.student):
            self.student = nn.SyncBatchNorm.convert_sync_batchnorm(self.student)
            self.teacher = nn.SyncBatchNorm.convert_sync_batchnorm(self.teacher)
        self.ddp_student = DDP(self.student)
        self.ddp_teacher = DDP(self.teacher)
        self.ddp_teacher.requires_grad_(False)  # type: ignore

    def init_with_trainer(self, trainer: ITrainer) -> None:
        self.teacher_for_training.requires_grad_(False)

    def permute_trainer_config(self, trainer_config: Dict[str, Any]) -> None:
        # TODO : make `permute_trainer_config` more general
        if trainer_config["clip_norm"] == 0.0:
            trainer_config["clip_norm"] = 3.0
        if trainer_config["lr"] is None:
            trainer_config["lr"] = 0.0005
        self.lr = trainer_config["lr"]
        self.min_lr = trainer_config.pop("min_lr", 1.0e-6)
        if trainer_config["optimizer_name"] is None:
            trainer_config["optimizer_name"] = "adamw"
        trainer_config["scheduler_name"] = "none"


__all__ = [
    "DINO",
]



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional
from cftool.types import tensor_dict_type

from .encoder import make_encoder
from .encoder import IEncoder
from .encoder import EncoderMixin
from .decoder import make_decoder
from .decoder import IDecoder
from ...misc.toolkit import auto_num_layers


class EncoderDecoderMixin:
    encoder: IEncoder
    decoder: IDecoder

    def resize(self, net: Tensor) -> Tensor:
        return self.decoder.resize(net)

    def encode(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
        return self.encoder.encode(batch, **kwargs)

    def decode(
        self,
        batch: tensor_dict_type,
        *,
        resize: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        net = self.decoder.decode(batch, **kwargs)
        if resize:
            net = self.resize(net)
        return net


class EncoderDecoder(nn.Module, EncoderDecoderMixin):
    def __init__(
        self,
        is_1d: bool,
        in_channels: int,
        out_channels: Optional[int] = None,
        target_downsample: Optional[int] = None,
        latent_padding_channels: Optional[int] = 16,
        num_classes: Optional[int] = None,
        *,
        latent: int = 128,
        decoder_latent: Optional[int] = None,
        img_size: Optional[int] = None,
        min_size: int = 2,
        num_downsample: Optional[int] = None,
        num_upsample: Optional[int] = None,
        latent_resolution: Optional[int] = None,
        encoder: str = "vanilla",
        decoder: str = "vanilla",
        encoder_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        # up / down sample stuffs
        if num_downsample is None:
            if img_size is None:
                raise ValueError(
                    "either `img_size` or `num_downsample` should be provided "
                    "(when `num_downsample` is not provided, it will be inferred "
                    "automatically with `img_size` & `min_size`)"
                )
            args = img_size, min_size, target_downsample
            num_downsample = auto_num_layers(*args, use_stride=encoder == "vanilla")
        if latent_resolution is None and img_size is None:
            raise ValueError(
                "either `img_size` or `latent_resolution` should be provided "
                "in `VanillaVAEBase` (when `latent_resolution` is not provided, "
                "it will be inferred automatically with `img_size`)"
            )
        # properties
        self.is_1d = is_1d
        self.latent = latent
        self.decoder_latent = decoder_latent or latent
        self.latent_key = "latent_dim" if is_1d else "latent_channels"
        self.img_size = img_size
        self.latent_padding_channels = latent_padding_channels
        self.num_classes = num_classes
        self.num_downsample = num_downsample
        self.num_upsample = num_upsample
        # encoder
        if encoder_config is None:
            encoder_config = {}
        encoder_config.setdefault("img_size", img_size)
        encoder_config.setdefault("in_channels", in_channels)
        encoder_config.setdefault("num_downsample", num_downsample)
        encoder_config[self.latent_key] = latent
        self.encoder = make_encoder(encoder, encoder_config, is_1d=is_1d)  # type: ignore
        if isinstance(self.encoder, EncoderMixin) and img_size is not None:
            latent_resolution = self.encoder.latent_resolution(img_size)
        # decoder
        if decoder_config is None:
            decoder_config = {}
        decoder_config.setdefault("img_size", img_size)
        decoder_config.setdefault("num_classes", num_classes)
        decoder_config.setdefault("out_channels", out_channels or in_channels)
        decoder_config.setdefault("num_upsample", num_upsample)
        decoder_config.setdefault("latent_resolution", latent_resolution)
        decoder_config[self.latent_key] = self.decoder_latent
        self.decoder = make_decoder(decoder, decoder_config, is_1d=is_1d)
        self.latent_resolution = self.decoder.latent_resolution


class PureEncoderDecoder(nn.Module, EncoderDecoderMixin):
    def __init__(
        self,
        *,
        is_1d: bool,
        encoder: str = "vanilla",
        decoder: str = "vanilla",
        encoder_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.encoder = make_encoder(encoder, encoder_config or {}, is_1d=is_1d)
        self.decoder = make_decoder(decoder, decoder_config or {}, is_1d=is_1d)


__all__ = [
    "EncoderDecoder",
    "PureEncoderDecoder",
]



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Tuple
from typing import Optional
from torch.fft import fft
from cftool.misc import safe_execute

from .protocols import TokenMixerBase
from .feedforward import FeedForward
from ..common import Lambda
from ..attentions import Attention


@TokenMixerBase.register("mlp")
class MLPTokenMixer(TokenMixerBase):
    def __init__(self, num_tokens: int, latent_dim: int, *, dropout: float = 0.1):
        super().__init__(num_tokens, latent_dim)
        self.net = nn.Sequential(
            Lambda(lambda x: x.transpose(1, 2), name="to_token_mixing"),
            FeedForward(num_tokens, num_tokens, dropout),
            Lambda(lambda x: x.transpose(1, 2), name="to_channel_mixing"),
        )

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
    ) -> Tensor:
        return self.net(net)


@TokenMixerBase.register("fourier")
class FourierTokenMixer(TokenMixerBase):
    def __init__(self, num_tokens: int, latent_dim: int):
        super().__init__(num_tokens, latent_dim)
        self.net = Lambda(lambda x: fft(fft(x, dim=-1), dim=-2).real, name="fourier")

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
    ) -> Tensor:
        return self.net(net)


@TokenMixerBase.register("attention")
class AttentionTokenMixer(TokenMixerBase):
    def __init__(
        self,
        num_tokens: int,
        latent_dim: int,
        *,
        attention_type: str = "basic",
        **attention_kwargs: Any,
    ):
        super().__init__(num_tokens, latent_dim)
        attention_kwargs.setdefault("bias", False)
        attention_kwargs.setdefault("num_heads", 8)
        attention_kwargs["input_dim"] = latent_dim
        attention_kwargs.setdefault("is_self_attention", True)
        self.net = safe_execute(Attention.get(attention_type), attention_kwargs)

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
        mask: Optional[Tensor] = None,
    ) -> Tensor:
        kw = dict(hw=hw, mask=mask, determinate=determinate)
        return self.net(net, net, net, **kw).output


@TokenMixerBase.register("pool")
class PoolTokenMixer(TokenMixerBase):
    def __init__(self, num_tokens: int, latent_dim: int, *, pool_size: int = 3):
        super().__init__(num_tokens, latent_dim)
        self.pool = nn.AvgPool2d(
            pool_size,
            stride=1,
            padding=pool_size // 2,
            count_include_pad=False,
        )

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
    ) -> Tensor:
        return self.pool(net) - net


__all__ = [
    "MLPTokenMixer",
    "FourierTokenMixer",
    "AttentionTokenMixer",
    "PoolTokenMixer",
]



import torch

import torch.nn as nn

from torch import Tensor
from typing import Tuple
from typing import Optional
from cftool.types import tensor_dict_type

from .common import IAutoEncoder
from .common import AutoEncoderInit
from .common import AutoEncoderLPIPSWithDiscriminator
from ..generator.vector_quantized import VQCodebook
from ....constants import INPUT_KEY
from ....constants import PREDICTIONS_KEY
from ....misc.internal_ import register_custom_module
from ....misc.internal_ import CustomTrainStepLoss
from ....losses.vae import VQVAELoss


class AutoEncoderVQ(AutoEncoderInit):
    enc_double_channels = False

    def __init__(
        self,
        img_size: int,
        in_channels: int,
        out_channels: int,
        inner_channels: int,
        latent_channels: int,
        channel_multipliers: Tuple[int, ...],
        *,
        num_code: int,
        embedding_channels: int,
        num_res_blocks: int,
        attention_resolutions: Tuple[int, ...] = (),
        dropout: float = 0.0,
        resample_with_conv: bool = True,
        attention_type: str = "spatial",
        apply_tanh: bool = False,
    ):
        super().__init__(
            img_size,
            in_channels,
            out_channels,
            inner_channels,
            latent_channels,
            channel_multipliers,
            embedding_channels=embedding_channels,
            num_res_blocks=num_res_blocks,
            attention_resolutions=attention_resolutions,
            dropout=dropout,
            resample_with_conv=resample_with_conv,
            attention_type=attention_type,
            apply_tanh=apply_tanh,
        )
        self.codebook = VQCodebook(num_code, embedding_channels)

    def encode(self, net: Tensor) -> Tensor:
        net = self.generator.encode({INPUT_KEY: net})
        net = self.to_embedding(net)
        return net

    def decode(
        self,
        z: Tensor,
        *,
        resize: bool = True,
        apply_codebook: bool = True,
        apply_tanh: Optional[bool] = None,
    ) -> Tensor:
        if apply_codebook:
            z = self.codebook(z).z_q
        net = self.from_embedding(z)
        kw = dict(resize=resize, apply_tanh=apply_tanh)
        net = self.generator.decode({INPUT_KEY: net}, **kw)  # type: ignore
        return net

    def forward(
        self,
        net: Tensor,
        *,
        apply_tanh: Optional[bool] = None,
    ) -> dict:
        net = self.encode(net)
        out = self.codebook(net, return_z_q_g=True)
        net = self.decode(out.z_q, apply_codebook=False, apply_tanh=apply_tanh)
        results = {PREDICTIONS_KEY: net}
        results.update(out.to_dict())
        return results


class AutoEncoderVQLoss(AutoEncoderLPIPSWithDiscriminator):
    def __init__(
        self,
        *,
        # vq
        lb_vq: float = 0.25,
        lb_recon: float = 1.0,
        lb_commit: float = 1.0,
        vq_loss_type: str = "l1",
        # common
        kl_weight: float = 1.0,
        d_loss: str = "hinge",
        d_loss_start_step: int = 50001,
        d_num_layers: int = 4,
        d_in_channels: int = 3,
        d_start_channels: int = 64,
        d_factor: float = 1.0,
        d_weight: float = 1.0,
        perceptual_weight: float = 1.0,
    ):
        super().__init__(
            kl_weight=kl_weight,
            d_loss=d_loss,
            d_loss_start_step=d_loss_start_step,
            d_num_layers=d_num_layers,
            d_in_channels=d_in_channels,
            d_start_channels=d_start_channels,
            d_factor=d_factor,
            d_weight=d_weight,
            perceptual_weight=perceptual_weight,
        )
        self.vq_loss = VQVAELoss(
            lb_vq=lb_vq,
            lb_recon=lb_recon,
            lb_commit=lb_commit,
            loss_type=vq_loss_type,
        )

    def get_generator_loss(
        self,
        batch: tensor_dict_type,
        forward_results: tensor_dict_type,
        *,
        step: int,
        last_layer: nn.Parameter,
        cond: Optional[Tensor] = None,
    ) -> CustomTrainStepLoss:
        inputs = batch[INPUT_KEY].contiguous()
        reconstructions = forward_results[PREDICTIONS_KEY].contiguous()
        # vq & nll loss
        vq_losses = self.vq_loss(forward_results, batch, reduction="none", gather=False)
        """{"mse": mse, "commit": commit_loss, LOSS_KEY: loss}"""
        recon_loss = vq_losses[self.vq_loss.loss_type]
        codebook_loss = vq_losses["codebook"].mean()
        with torch.no_grad():
            loss_items = {
                "recon": recon_loss.mean().item(),
                "codebook": codebook_loss.item(),
            }
        if self.perceptual_weight > 0:
            p_loss = self.perceptual_loss(inputs, reconstructions)
            p_loss = self.perceptual_weight * p_loss
            recon_loss = recon_loss + p_loss
            with torch.no_grad():
                loss_items["perceptual"] = p_loss.mean().item()
        nll_loss = torch.mean(recon_loss)
        loss = nll_loss + codebook_loss
        # check discriminator start
        if step < self.d_loss_start_step:
            return CustomTrainStepLoss(loss, loss_items)
        g_loss = self.g_loss(nll_loss, last_layer, loss_items, reconstructions, cond)
        loss = loss + g_loss
        return CustomTrainStepLoss(loss, loss_items)


@register_custom_module("ae_vq")
class AutoEncoderVQModel(AutoEncoderVQ, IAutoEncoder):  # type: ignore
    def __init__(
        self,
        img_size: int,
        in_channels: int,
        out_channels: int,
        inner_channels: int,
        latent_channels: int,
        channel_multipliers: Tuple[int, ...],
        *,
        num_code: int,
        embedding_channels: int,
        num_res_blocks: int,
        attention_resolutions: Tuple[int, ...] = (),
        dropout: float = 0.0,
        resample_with_conv: bool = True,
        attention_type: str = "spatial",
        # loss configs
        use_loss: bool = True,
        lb_vq: float = 0.25,
        lb_recon: float = 1.0,
        lb_commit: float = 1.0,
        vq_loss_type: str = "l1",
        kl_weight: float = 1.0,
        d_loss: str = "hinge",
        d_loss_start_step: int = 50001,
        d_num_layers: int = 4,
        d_in_channels: int = 3,
        d_start_channels: int = 64,
        d_factor: float = 1.0,
        d_weight: float = 1.0,
        perceptual_weight: float = 1.0,
        # train step configs
        grad_accumulate: int = 1,
    ):
        super().__init__(
            img_size,
            in_channels,
            out_channels,
            inner_channels,
            latent_channels,
            channel_multipliers,
            num_code=num_code,
            embedding_channels=embedding_channels,
            num_res_blocks=num_res_blocks,
            attention_resolutions=attention_resolutions,
            dropout=dropout,
            resample_with_conv=resample_with_conv,
            attention_type=attention_type,
        )
        if not use_loss:
            self.loss = None
        else:
            self.loss = AutoEncoderVQLoss(
                lb_vq=lb_vq,
                lb_recon=lb_recon,
                lb_commit=lb_commit,
                vq_loss_type=vq_loss_type,
                kl_weight=kl_weight,
                d_loss=d_loss,
                d_loss_start_step=d_loss_start_step,
                d_num_layers=d_num_layers,
                d_in_channels=d_in_channels,
                d_start_channels=d_start_channels,
                d_factor=d_factor,
                d_weight=d_weight,
                perceptual_weight=perceptual_weight,
            )
        self.setup(img_size, grad_accumulate, embedding_channels, channel_multipliers)


__all__ = [
    "AutoEncoderVQ",
]



import torch

import numpy as np

from abc import abstractmethod
from abc import ABCMeta

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Protocol

from .utils import append_dims
from .utils import append_zero
from .protocol import ISampler
from .protocol import IQSampler
from .protocol import IDiffusion
from .protocol import UncondSamplerMixin
from ..utils import extract_to

try:
    from scipy import integrate
except:
    integrate = None


def to_d(image: Tensor, sigma: Tensor, denoised: Tensor) -> Tensor:
    return (image - denoised) / append_dims(sigma, image.ndim)


class IGetDenoised(Protocol):
    def __call__(self, img: Tensor) -> Tensor:
        pass


class KQSampler(IQSampler):
    def q_sample(
        self,
        net: Tensor,
        timesteps: Tensor,
        noise: Optional[Tensor] = None,
    ) -> Tensor:
        w_noise = extract_to(self.sigmas, timesteps, net.ndim)
        if noise is None:
            noise = torch.randn_like(net)
        net = net + w_noise * noise
        return net

    def reset_buffers(self, sigmas: Tensor) -> None:  # type: ignore
        self.sigmas = sigmas


class KSamplerMixin(ISampler, UncondSamplerMixin, metaclass=ABCMeta):
    def __init__(
        self,
        model: IDiffusion,
        *,
        default_quantize: bool = False,
        unconditional_cond: Optional[Any] = None,
        unconditional_guidance_scale: float = 1.0,
        default_steps: int = 50,
    ):
        if model.parameterization != "eps":
            raise ValueError("only `eps` parameterization is supported in `k_samplers`")
        super().__init__(model)
        self.default_quantize = default_quantize
        self.unconditional_cond = unconditional_cond
        self.unconditional_guidance_scale = unconditional_guidance_scale
        self.default_steps = default_steps

    # abstract

    @abstractmethod
    def sample_step_core(
        self,
        image: Tensor,
        cond: Optional[Tensor],
        step: int,
        total_step: int,
        get_denoised: IGetDenoised,
        *,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        **kwargs: Any,
    ) -> Tensor:
        pass

    # inheritance

    @property
    def q_sampler(self) -> KQSampler:
        return self._q_sampler

    @property
    def sample_kwargs(self) -> Dict[str, Any]:
        return dict(
            quantize=self.default_quantize,
            unconditional_cond=self.unconditional_cond,
            unconditional_guidance_scale=self.unconditional_guidance_scale,
        )

    def sample_step(
        self,
        image: Tensor,
        cond: Optional[Tensor],
        step: int,
        total_step: int,
        *,
        quantize: bool = False,
        unconditional_cond: Optional[Any] = None,
        unconditional_guidance_scale: float = 1.0,
        **kwargs: Any,
    ) -> Tensor:
        if step == 0:
            self._reset_buffers(
                total_step,
                quantize,
                unconditional_cond,
                unconditional_guidance_scale,
            )
            image = image * self.sigmas[0]
        if step == total_step - 1:
            return image

        def get_denoised(img: Tensor) -> Tensor:
            sigmas = self.sigmas[step] * image.new_ones([image.shape[0]])
            ts = self._sigma_to_t(sigmas, quantize)
            c_in = append_dims(
                1.0 / (sigmas**2 + self.sigma_data**2) ** 0.5, img.ndim
            )
            c_out = append_dims(sigmas, img.ndim)
            eps = self._uncond_denoise(img * c_in, ts, cond)
            return img - eps * c_out

        return self.sample_step_core(
            image,
            cond,
            step,
            total_step,
            get_denoised,
            quantize=quantize,
            unconditional_cond=unconditional_cond,
            unconditional_guidance_scale=unconditional_guidance_scale,
            **kwargs,
        )

    # internal

    def _t_to_sigma(self, ts: Tensor) -> Tensor:
        low_idx, high_idx, w = ts.floor().long(), ts.ceil().long(), ts.frac()
        return (1.0 - w) * self.sigmas_base[low_idx] + w * self.sigmas_base[high_idx]

    def _sigma_to_t(self, sigmas: Tensor, quantize: bool) -> Tensor:
        dists = torch.abs(sigmas - self.sigmas_base[:, None])
        if quantize:
            return torch.argmin(dists, dim=0).view(sigmas.shape)
        top2 = torch.topk(dists, dim=0, k=2, largest=False).indices
        low_idx, high_idx = torch.sort(top2, dim=0)[0]
        low, high = self.sigmas_base[low_idx], self.sigmas_base[high_idx]
        w = (low - sigmas) / (low - high)
        w = w.clamp(0, 1)
        ts = (1.0 - w) * low_idx + w * high_idx
        return ts.view(sigmas.shape)

    def _reset_buffers(
        self,
        total_step: int,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
    ) -> None:
        alphas = self.model.alphas_cumprod
        self.sigmas_base = ((1.0 - alphas) / alphas) ** 0.5
        t_max = len(self.sigmas_base) - 1
        ts = torch.linspace(t_max, 0, total_step, device=self.sigmas_base.device)
        self.sigmas = append_zero(self._t_to_sigma(ts)).to(alphas.dtype)
        self.sigma_data = 1.0
        self.quantize = quantize
        # q sampling
        self._q_sampler = KQSampler(self.model)
        self._q_sampler.reset_buffers(self.sigmas.flip(0))
        # unconditional conditioning
        self._reset_uncond_buffers(unconditional_cond, unconditional_guidance_scale)


def klms_coef(order: int, t: np.ndarray, i: int, j: int) -> float:
    def fn(tau: float) -> float:
        prod = 1.0
        for k in range(order):
            if j == k:
                continue
            prod *= (tau - t[i - k]) / (t[i - j] - t[i - k])
        return prod

    if integrate is None:
        raise ValueError("`scipy` is needed for `KLMSSampler`")
    return integrate.quad(fn, t[i], t[i + 1], epsrel=1.0e-4)[0]


@ISampler.register("klms")
class KLMSSampler(KSamplerMixin):
    def sample_step_core(
        self,
        image: Tensor,
        cond: Optional[Tensor],
        step: int,
        total_step: int,
        get_denoised: IGetDenoised,
        *,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        **kwargs: Any,
    ) -> Tensor:
        order = 4
        denoised = get_denoised(image)
        d = to_d(image, self.sigmas[step], denoised)
        self.ds.append(d)
        if len(self.ds) > order:
            self.ds.pop(0)
        current_order = min(step + 1, order)
        for i, d in enumerate(reversed(self.ds)):
            coef = klms_coef(current_order, self.sigmas_numpy, step, i)
            image = image + coef * d
        return image

    def _reset_buffers(
        self,
        total_step: int,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
    ) -> None:
        super()._reset_buffers(
            total_step,
            quantize,
            unconditional_cond,
            unconditional_guidance_scale,
        )
        self.ds: List[Tensor] = []
        self.sigmas_numpy = self.sigmas.detach().cpu().numpy()


__all__ = [
    "KLMSSampler",
]



import numpy as np

from typing import List
from typing import NamedTuple


class ImageFolderLatentResponse(NamedTuple):
    latent: np.ndarray
    img_paths: List[str]


class IImageExtractor:
    def get_folder_latent(
        self,
        src_folder: str,
        *,
        batch_size: int,
        num_workers: int = 0,
        use_tqdm: bool = True,
    ) -> ImageFolderLatentResponse:
        pass


__all__ = [
    "IImageExtractor",
]



import os
import torch

import numpy as np

from cftool.array import to_numpy
from cftool.array import to_torch
from cftool.array import to_device
from cftool.array import save_images

from .general import ImageCallback
from ...toolkit import eval_context
from ....protocol import _forward
from ....protocol import ITrainer
from ....protocol import TrainerCallback
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import STYLE_KEY
from ....constants import PREDICTIONS_KEY
from ....models.protocols.cv import GeneratorMixin
from ....models.cv.segmentor.constants import LV1_ALPHA_KEY

try:
    from cfcv.misc.toolkit import min_max_normalize
except:
    min_max_normalize = None


@TrainerCallback.register("ldm")
@TrainerCallback.register("ddpm")
@TrainerCallback.register("ae_kl")
@TrainerCallback.register("ae_vq")
@TrainerCallback.register("gan")
@TrainerCallback.register("vae")
@TrainerCallback.register("vae2d")
@TrainerCallback.register("style_vae")
@TrainerCallback.register("generator")
class GeneratorCallback(ImageCallback):
    def __init__(self, num_keep: int = 25, num_interpolations: int = 16):
        super().__init__(num_keep)
        self.num_interpolations = num_interpolations

    def log_artifacts(self, trainer: ITrainer) -> None:
        if not self.is_rank_0:
            return None
        batch = next(iter(trainer.validation_loader))
        batch = to_device(batch, trainer.device)
        original = batch[INPUT_KEY]
        model = trainer.model.core
        if not isinstance(model, GeneratorMixin):
            msg = "`GeneratorCallback` is only compatible with `GeneratorMixin`"
            raise ValueError(msg)
        is_conditional = getattr(model, "is_conditional", False)
        labels = None if not is_conditional else batch[LABEL_KEY]
        image_folder = self._prepare_folder(trainer)
        # original
        save_images(original, os.path.join(image_folder, "original.png"))
        # reconstruct
        if model.can_reconstruct:
            with eval_context(model):
                reconstructed = model.reconstruct(original, labels=labels)
            save_images(reconstructed, os.path.join(image_folder, "reconstructed.png"))
        # sample
        with eval_context(model):
            sampled = model.sample(len(original))
        save_images(sampled, os.path.join(image_folder, "sampled.png"))
        # interpolation
        with eval_context(model):
            interpolations = model.interpolate(self.num_interpolations)
        save_images(interpolations, os.path.join(image_folder, "interpolations.png"))
        # conditional sampling
        if getattr(model, "num_classes", None) is None:
            return None
        cond_folder = os.path.join(image_folder, "conditional")
        os.makedirs(cond_folder, exist_ok=True)
        with eval_context(model):
            for i in range(model.num_classes or 0):
                sampled = model.sample(len(original), class_idx=i)
                interpolations = model.interpolate(len(original), class_idx=i)
                save_images(sampled, os.path.join(cond_folder, f"sampled_{i}.png"))
                path = os.path.join(cond_folder, f"interpolations_{i}.png")
                save_images(interpolations, path)


@TrainerCallback.register("siren_gan")
@TrainerCallback.register("siren_vae")
@TrainerCallback.register("sized_generator")
class SizedGeneratorCallback(GeneratorCallback):
    def log_artifacts(self, trainer: ITrainer) -> None:
        if not self.is_rank_0:
            return None
        super().log_artifacts(trainer)
        model = trainer.model.core
        image_folder = self._prepare_folder(trainer, check_num_keep=False)
        sample_method = getattr(model, "sample", None)
        if sample_method is None:
            raise ValueError(
                "`sample` should be implemented when `SizedGeneratorCallback` is used "
                "(and the `sample` method should support accepting `size` kwarg)"
            )
        resolution = getattr(model, "img_size", 32)
        with eval_context(model):
            for i in range(1, 4):
                size = resolution * 2**i
                batch_size = 1 if size > 256 else 4
                sampled = sample_method(batch_size, size=size).cpu()
                path = os.path.join(image_folder, f"sampled_{size}x{size}.png")
                save_images(sampled, path)


@ImageCallback.register("unet")
@ImageCallback.register("u2net")
@ImageCallback.register("cascade_u2net")
class AlphaSegmentationCallback(ImageCallback):
    def log_artifacts(self, trainer: ITrainer) -> None:
        if not self.is_rank_0:
            return None
        batch = next(iter(trainer.validation_loader))
        batch = to_device(batch, trainer.device)
        model = trainer.model.core
        with eval_context(model):
            results = _forward(model, 0, batch, INPUT_KEY)
        original = batch[INPUT_KEY]
        label = batch[LABEL_KEY].float()
        logits = results[PREDICTIONS_KEY]
        if isinstance(logits, list):
            logits = logits[0]
        if logits.shape[1] != 1:
            logits = logits[:, [-1]]
        seg_map = torch.sigmoid(logits)
        if min_max_normalize is not None:
            seg_map = min_max_normalize(seg_map)
        sharp_map = (seg_map > 0.5).to(torch.float32)
        image_folder = self._prepare_folder(trainer)
        save_images(original, os.path.join(image_folder, "original.png"))
        save_images(label, os.path.join(image_folder, "label.png"))
        save_images(seg_map, os.path.join(image_folder, "mask.png"))
        save_images(sharp_map, os.path.join(image_folder, "mask_sharp.png"))
        lv1_alpha = results.get(LV1_ALPHA_KEY)
        if lv1_alpha is not None:
            save_images(lv1_alpha, os.path.join(image_folder, "lv1_mask.png"))
        np_original = to_numpy(original)
        if min_max_normalize is not None:
            np_original = min_max_normalize(np_original)
        np_label, np_mask, np_sharp = map(to_numpy, [label, seg_map, sharp_map])
        rgba = np.concatenate([np_original, np_label], axis=1)
        rgba_pred = np.concatenate([np_original, np_mask], axis=1)
        rgba_sharp = np.concatenate([np_original, np_sharp], axis=1)
        save_images(to_torch(rgba), os.path.join(image_folder, "rgba.png"))
        save_images(to_torch(rgba_pred), os.path.join(image_folder, "rgba_pred.png"))
        save_images(to_torch(rgba_sharp), os.path.join(image_folder, "rgba_sharp.png"))


@TrainerCallback.register("adain")
@TrainerCallback.register("style_transfer")
class StyleTransferCallback(ImageCallback):
    def log_artifacts(self, trainer: ITrainer) -> None:
        if not self.is_rank_0:
            return None
        batch = next(iter(trainer.validation_loader))
        batch = to_device(batch, trainer.device)
        content = batch[INPUT_KEY]
        style = batch[STYLE_KEY]
        model = trainer.model.core
        image_folder = self._prepare_folder(trainer)
        # inputs
        save_images(content, os.path.join(image_folder, "original.png"))
        save_images(style, os.path.join(image_folder, "styles.png"))
        # stylize
        with eval_context(model):
            stylized = model.stylize(content, style)
        save_images(stylized, os.path.join(image_folder, "stylized.png"))


__all__ = [
    "GeneratorCallback",
    "SizedGeneratorCallback",
    "AlphaSegmentationCallback",
    "StyleTransferCallback",
]



from typing import Any
from typing import Optional
from cftool.types import np_dict_type
from cftool.types import tensor_dict_type

from .core import HuggingFaceModel
from ....types import texts_type
from ....protocol import TrainerState
from ....constants import PREDICTIONS_KEY

try:
    from transformers import MarianMTModel
    from transformers import MarianTokenizer
except:
    MarianMTModel = MarianTokenizer = None


@HuggingFaceModel.register("opus-base")
class OPUSBase(HuggingFaceModel):
    model_base = MarianMTModel
    tokenizer_base = MarianTokenizer
    forward_fn_name = "generate"

    def __init__(self, src: str, tgt: str) -> None:
        if MarianMTModel is None:
            raise ValueError("`transformers` is needed for `OPUSBase`")
        super().__init__(f"Helsinki-NLP/opus-mt-{src}-{tgt}")

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional["TrainerState"] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        generated_ids = super().model_forward(batch)
        return {PREDICTIONS_KEY: generated_ids}

    def inference(
        self,
        texts: texts_type,
        use_tqdm: bool = True,
        **kwargs: Any,
    ) -> np_dict_type:
        return super().inference(texts, use_tqdm, stack_outputs=False, **kwargs)


@HuggingFaceModel.register("opus-zh-en")
class OPUS_ZH_EN(OPUSBase):
    def __init__(self) -> None:
        super().__init__("zh", "en")


__all__ = [
    "OPUSBase",
    "OPUS_ZH_EN",
]



from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional
from cftool.types import tensor_dict_type

from ..bases import IBAKE
from ..bases import IRDropout
from ..register import register_ml_module
from ..register import register_custom_loss_module
from ...constants import LATENT_KEY
from ...constants import PREDICTIONS_KEY
from ..protocols.ml import MixedStackedModel


@register_ml_module("mixer")
class Mixer(MixedStackedModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int,
        latent_dim: int = 256,
        *,
        num_layers: int = 4,
        dropout: float = 0.0,
        norm_type: Optional[str] = "batch_norm",
    ):
        super().__init__(
            input_dim,
            output_dim,
            num_history,
            latent_dim,
            token_mixing_type="mlp",
            num_layers=num_layers,
            dropout=dropout,
            norm_type=norm_type,
            feedforward_dim_ratio=1.0,
            use_head_token=False,
            use_positional_encoding=False,
        )


@register_custom_loss_module("mixer_bake", is_ml=True)
class MixerWithBAKE(IBAKE):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int,
        latent_dim: int = 256,
        *,
        num_layers: int = 4,
        dropout: float = 0.1,
        norm_type: Optional[str] = "batch_norm",
        lb: float = 0.1,
        bake_loss: str = "auto",
        bake_loss_config: Optional[Dict[str, Any]] = None,
        w_ensemble: float = 0.5,
        is_classification: bool,
    ):
        super().__init__()
        self.mixer = Mixer(
            input_dim,
            output_dim,
            num_history,
            latent_dim,
            num_layers=num_layers,
            dropout=dropout,
            norm_type=norm_type,
        )
        self._init_bake(lb, bake_loss, bake_loss_config, w_ensemble, is_classification)

    def forward(self, net: Tensor) -> tensor_dict_type:  # type: ignore
        net = self.mixer.to_encoder(net)
        latent = self.mixer.encoder(net)
        net = self.mixer.head(latent)
        return {LATENT_KEY: latent, PREDICTIONS_KEY: net}


@register_custom_loss_module("mixer_r_dropout", is_ml=True)
class MixerWithRDropout(IRDropout):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int,
        latent_dim: int = 256,
        *,
        num_layers: int = 4,
        dropout: float = 0.1,
        norm_type: Optional[str] = "batch_norm",
        lb: float = 0.1,
        is_classification: bool,
    ):
        super().__init__()
        self.mixer = Mixer(
            input_dim,
            output_dim,
            num_history,
            latent_dim,
            num_layers=num_layers,
            dropout=dropout,
            norm_type=norm_type,
        )
        # R-Dropout
        self.lb = lb
        self.is_classification = is_classification

    def forward(self, net: Tensor) -> tensor_dict_type:  # type: ignore
        return self.mixer(net)


__all__ = [
    "Mixer",
    "MixerWithBAKE",
    "MixerWithRDropout",
]



import math
import torch

import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Optional
from torch.nn import Module
from torch.nn import ModuleList
from cftool.misc import safe_execute

from .poolers import BertPooler
from .poolers import SequencePooler
from .feedforward import FFN
from .feedforward import FeedForward
from .token_mixers import TokenMixerBase
from ..norms import NormFactory
from ..utils import zero_module
from ..common import Lambda
from ..customs import Linear
from ..customs import DropPath
from ..high_level import PreNorm
from ..attentions import CrossAttention
from ....misc.toolkit import interpolate
from ....misc.toolkit import gradient_checkpoint


class MixingBlock(Module):
    def __init__(
        self,
        num_tokens: int,
        latent_dim: int,
        feedforward_dim: int,
        *,
        norm_position: str = "pre_norm",
        token_mixing_type: str,
        token_mixing_config: Optional[Dict[str, Any]] = None,
        token_mixing_dropout: Optional[float] = None,
        channel_mixing_type: str = "ff",
        channel_mixing_config: Optional[Dict[str, Any]] = None,
        dropout: float = 0.0,
        drop_path: float = 0.0,
        norm_type: Optional[str] = "batch_norm",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        residual_after_norm: bool = False,
    ):
        def _make_norm() -> nn.Module:
            factory = NormFactory(norm_type)
            return factory.make(latent_dim, **(norm_kwargs or {}))

        super().__init__()
        if norm_position not in {"pre_norm", "post_norm"}:
            raise ValueError(
                "`norm_position` should be either 'pre_norm' or 'post_norm', "
                f"'{norm_position}' found"
            )
        self.norm_position = norm_position
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        # token mixing
        if token_mixing_config is None:
            token_mixing_config = {}
        token_mixing_config.update(
            {
                "num_tokens": num_tokens,
                "latent_dim": latent_dim,
                "dropout": dropout,
            }
        )
        self.token_norm = _make_norm()
        token_mixing_base = TokenMixerBase.get(token_mixing_type)
        self.token_mixing = safe_execute(token_mixing_base, token_mixing_config)
        if token_mixing_dropout is None:
            token_mixing_dropout = dropout
        self.token_mixing_dropout = nn.Dropout(token_mixing_dropout)
        # channel mixing
        if channel_mixing_config is None:
            channel_mixing_config = {}
        channel_mixing_config.update(
            {
                "in_dim": latent_dim,
                "latent_dim": feedforward_dim,
                "dropout": dropout,
            }
        )
        self.residual_after_norm = residual_after_norm
        self.channel_norm = _make_norm()
        self.channel_mixing = FFN.make(channel_mixing_type, channel_mixing_config)

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        if self.norm_position == "pre_norm":
            return self._pre_norm_forward(net, hw, determinate=determinate, **kwargs)
        if self.norm_position == "post_norm":
            return self._post_norm_forward(net, hw, determinate=determinate, **kwargs)
        raise ValueError(f"unrecognized norm_position '{self.norm_position}' occurred")

    def _pre_norm_forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        # token mixing
        token_mixing_kw = dict(hw=hw, determinate=determinate)
        token_mixing_kw.update(kwargs)
        token_mixing_net = self.token_norm(net)
        token_mixing_net = self.token_mixing(token_mixing_net, **token_mixing_kw)
        token_mixing_net = self.token_mixing_dropout(token_mixing_net)
        net = net + self.drop_path(token_mixing_net)
        # channel mixing
        if self.residual_after_norm:
            net = self.channel_norm(net)
        if not self.channel_mixing.need_2d:
            channel_mixing_net = net
        else:
            if hw is None:
                raise ValueError("`hw` should be provided when FFN needs 2d input")
            channel_mixing_net = net.view(-1, *hw, net.shape[-1])
        if not self.residual_after_norm:
            channel_mixing_net = self.channel_norm(channel_mixing_net)
        channel_mixing_net = self.channel_mixing(channel_mixing_net)
        net = net + self.drop_path(channel_mixing_net)
        return net

    def _post_norm_forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        # token mixing
        token_mixing_kw = dict(hw=hw, determinate=determinate)
        token_mixing_kw.update(kwargs)
        token_mixing_net = self.token_mixing(net, **token_mixing_kw)
        token_mixing_net = self.token_mixing_dropout(token_mixing_net)
        net = net + self.drop_path(token_mixing_net)
        net = self.token_norm(net)
        # channel mixing
        if not self.channel_mixing.need_2d:
            channel_mixing_net = net
        else:
            if hw is None:
                raise ValueError("`hw` should be provided when FFN needs 2d input")
            channel_mixing_net = net.view(-1, *hw, net.shape[-1])
        channel_mixing_net = self.channel_mixing(channel_mixing_net)
        net = net + self.drop_path(channel_mixing_net)
        net = self.channel_norm(net)
        return net


class PositionalEncoding(Module):
    def __init__(
        self,
        dim: int,
        num_history: int,
        dropout: float = 0.0,
        *,
        num_heads: int,
        is_vision: bool,
        enable: bool = True,
    ):
        super().__init__()
        self.pos_drop = None
        self.pos_encoding = None
        if enable:
            self.pos_drop = nn.Dropout(p=dropout)
            self.pos_encoding = nn.Parameter(torch.zeros(1, num_history, dim))
            nn.init.trunc_normal_(self.pos_encoding, std=0.02)
        self.num_heads = num_heads
        self.is_vision = is_vision

    def forward(
        self,
        net: Tensor,
        *,
        hwp: Optional[Tuple[int, int, int]] = None,
    ) -> Tensor:
        if self.pos_encoding is None or self.pos_drop is None:
            return net
        if self.is_vision:
            pos_encoding = self.interpolate_pos_encoding(net, hwp)
        else:
            pos_encoding = self.pos_encoding[:, : net.shape[1] - self.num_heads]
        pos_encoding = self.pos_drop(pos_encoding)
        return net + pos_encoding

    # this is for vision positional encodings
    def interpolate_pos_encoding(
        self,
        net: Tensor,
        hwp: Optional[Tuple[int, int, int]],
    ) -> Tensor:
        pos_encoding = self.pos_encoding
        assert pos_encoding is not None
        num_current_history = net.shape[1] - self.num_heads
        num_history = pos_encoding.shape[1] - self.num_heads
        if hwp is None:
            w = h = patch_size = None
        else:
            h, w, patch_size = hwp
        if num_current_history == num_history and w == h:
            return pos_encoding
        if w is None or h is None or patch_size is None:
            raise ValueError("`hwp` should be provided for `interpolate_pos_encoding`")
        head_encoding = None
        if self.num_heads > 0:
            head_encoding = pos_encoding[:, : self.num_heads]
            pos_encoding = pos_encoding[:, self.num_heads :]
        dim = net.shape[-1]
        # This assume that the original input is squared image
        sqrt = math.sqrt(num_history)
        wh_ratio = w / h
        pw = math.sqrt(num_current_history * wh_ratio) + 0.1
        ph = math.sqrt(num_current_history / wh_ratio) + 0.1
        pos_encoding = interpolate(
            pos_encoding.reshape(1, int(sqrt), int(sqrt), dim).permute(0, 3, 1, 2),
            factor=(pw / sqrt, ph / sqrt),
            mode="bicubic",
        )
        assert int(pw) == pos_encoding.shape[-2] and int(ph) == pos_encoding.shape[-1]
        pos_encoding = pos_encoding.permute(0, 2, 3, 1).view(1, -1, dim)
        if head_encoding is None:
            return pos_encoding
        return torch.cat([head_encoding, pos_encoding], dim=1)


class MixedStackedEncoder(Module):
    def __init__(
        self,
        dim: int,
        num_history: int,
        *,
        token_mixing_type: str,
        token_mixing_config: Optional[Dict[str, Any]] = None,
        channel_mixing_type: str = "ff",
        channel_mixing_config: Optional[Dict[str, Any]] = None,
        num_layers: int = 4,
        dropout: float = 0.0,
        dpr_list: Optional[List[float]] = None,
        drop_path_rate: float = 0.1,
        norm_position: str = "pre_norm",
        norm_type: Optional[str] = "batch_norm",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        embedding_norm: Optional[nn.Module] = None,
        embedding_dropout: Optional[float] = None,
        residual_after_norm: bool = False,
        feedforward_dim_ratio: float = 1.0,
        use_head_token: bool = False,
        head_pooler: Optional[str] = "mean",
        use_positional_encoding: bool = False,
        is_vision_positional_encoding: Optional[bool] = None,
        positional_encoding_dropout: float = 0.0,
        no_head_norm: Optional[bool] = None,
        norm_after_head: bool = False,
        aux_heads: Optional[List[str]] = None,
    ):
        super().__init__()
        # head token
        self.aux_heads = aux_heads
        if not use_head_token:
            num_heads = 0
            self.head_token = None
        else:
            num_heads = 1
            if aux_heads is not None:
                num_heads += len(aux_heads)
            self.head_token = nn.Parameter(torch.zeros(1, num_heads, dim))
        self.num_heads = num_heads
        # positional encoding
        num_history += num_heads
        if is_vision_positional_encoding is None:
            if use_positional_encoding:
                raise ValueError(
                    "`is_vision_positional_encoding` should be specified when "
                    "`is_vision_positional_encoding` is set to True"
                )
            is_vision_positional_encoding = False
        self.pos_encoding = PositionalEncoding(
            dim,
            num_history,
            positional_encoding_dropout,
            num_heads=num_heads,
            is_vision=is_vision_positional_encoding,
            enable=use_positional_encoding,
        )
        self.embedding_norm = embedding_norm
        if embedding_dropout is None:
            self.embedding_dropout = None
        else:
            self.embedding_dropout = nn.Dropout(embedding_dropout)
        # core
        if dpr_list is None:
            dpr_list = [x.item() for x in torch.linspace(0, drop_path_rate, num_layers)]
        feedforward_dim = int(round(dim * feedforward_dim_ratio))
        self.mixing_blocks = ModuleList(
            [
                MixingBlock(
                    num_history,
                    dim,
                    feedforward_dim,
                    norm_position=norm_position,
                    token_mixing_type=token_mixing_type,
                    token_mixing_config=token_mixing_config,
                    channel_mixing_type=channel_mixing_type,
                    channel_mixing_config=channel_mixing_config,
                    dropout=dropout,
                    drop_path=drop_path,
                    norm_type=norm_type,
                    norm_kwargs=norm_kwargs,
                    residual_after_norm=residual_after_norm,
                )
                for drop_path in dpr_list
            ]
        )
        # head
        if self.head_token is not None:
            if self.aux_heads is None:
                head = Lambda(lambda x: x[:, 0], name="head_token")
            else:
                head = Lambda(lambda x: x[:, : self.num_heads], name="head_token")
        else:
            if head_pooler is None:
                head = nn.Identity()
            elif head_pooler == "sequence":
                head = SequencePooler(dim, aux_heads)
            else:
                if aux_heads is not None:
                    raise ValueError(
                        "either `head_token` or `sequence` head_pooler should be used "
                        f"when `aux_heads` ({aux_heads}) is provided"
                    )
                if head_pooler == "bert":
                    head = BertPooler(dim)
                elif head_pooler == "mean":
                    head = Lambda(lambda x: x.mean(1), name="global_average")
                else:
                    msg = f"unrecognized head_pooler '{head_pooler}' occurred"
                    raise ValueError(msg)
        # head norm
        if no_head_norm is None:
            no_head_norm = norm_position == "post_norm"
        if no_head_norm:
            self.head_norm = None
            self.head = head
        elif norm_after_head:
            self.head_norm = NormFactory(norm_type).make(dim, **(norm_kwargs or {}))
            self.head = head
        else:
            self.head_norm = None
            self.head = PreNorm(
                dim,
                module=head,
                norm_type=norm_type,
                norm_kwargs=norm_kwargs,
            )
        # initializations
        if self.head_token is not None:
            nn.init.trunc_normal_(self.head_token, std=0.02)
        self.apply(self._init_weights)

    def _init_weights(self, m: Module) -> None:
        if isinstance(m, nn.Linear) or isinstance(m, Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0.0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0.0)
            nn.init.constant_(m.weight, 1.0)

    def pre_process(
        self,
        net: Tensor,
        *,
        determinate: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        n, t, d = net.shape
        if self.head_token is not None:
            head_tokens = self.head_token.repeat([n, 1, 1])
            net = torch.cat([head_tokens, net], dim=1)
            t += 1
        if determinate:
            net = net.view(-1, *map(int, [t, d]))
        net = self.pos_encoding(net, **kwargs)
        if self.embedding_norm is not None:
            net = self.embedding_norm(net)
        if self.embedding_dropout is not None:
            net = self.embedding_dropout(net)
        return net

    def post_process(self, net: Tensor) -> Tensor:
        net = self.head(net)
        if self.head_norm is not None:
            net = self.head_norm(net)
        return net

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        **kwargs: Any,
    ) -> Tensor:
        determinate = kwargs.pop("determinate", False)
        net = self.pre_process(net, determinate=determinate, **kwargs)
        for block in self.mixing_blocks:
            net = block(net, hw, determinate=determinate)
        net = self.post_process(net)
        return net


class SpatialTransformerBlock(Module):
    def __init__(
        self,
        query_dim: int,
        num_heads: int,
        head_dim: int,
        *,
        dropout: float = 0.0,
        context_dim: Optional[int] = None,
        feedforward_multiplier: float = 4.0,
        feedforward_activation: str = "geglu",
        use_checkpoint: bool = False,
        attn_split_chunk: Optional[int] = None,
    ):
        super().__init__()
        self.attn1 = CrossAttention(
            query_dim=query_dim,
            num_heads=num_heads,
            head_dim=head_dim,
            dropout=dropout,
            attn_split_chunk=attn_split_chunk,
        )
        latent_dim = round(query_dim * feedforward_multiplier)
        self.ff = FeedForward(
            query_dim,
            latent_dim,
            dropout,
            activation=feedforward_activation,
            add_last_dropout=False,
        )
        self.attn2 = CrossAttention(
            query_dim=query_dim,
            context_dim=context_dim,
            num_heads=num_heads,
            head_dim=head_dim,
            dropout=dropout,
            attn_split_chunk=attn_split_chunk,
        )
        self.norm1 = nn.LayerNorm(query_dim)
        self.norm2 = nn.LayerNorm(query_dim)
        self.norm3 = nn.LayerNorm(query_dim)
        self.use_checkpoint = use_checkpoint

    def forward(self, net: Tensor, context: Optional[Tensor] = None) -> Tensor:
        inputs = (net,) if context is None else (net, context)
        return gradient_checkpoint(
            self._forward,
            inputs=inputs,
            params=self.parameters(),
            enabled=self.use_checkpoint,
        )

    def _forward(self, net: Tensor, context: Optional[Tensor] = None) -> Tensor:
        net = self.attn1(self.norm1(net)) + net
        net = self.attn2(self.norm2(net), context=context) + net
        net = self.ff(self.norm3(net)) + net
        return net


class SpatialTransformer(Module):
    def __init__(
        self,
        in_channels: int,
        num_heads: int,
        head_dim: int,
        *,
        num_layers: int = 1,
        dropout: float = 0.0,
        context_dim: Optional[int] = None,
        use_checkpoint: bool = False,
        attn_split_chunk: Optional[int] = None,
    ):
        super().__init__()
        self.norm = nn.GroupNorm(32, in_channels, 1.0e-6, affine=True)
        latent_channels = num_heads * head_dim
        self.to_latent = nn.Conv2d(in_channels, latent_channels, 1, 1, 0)
        self.blocks = nn.ModuleList(
            [
                SpatialTransformerBlock(
                    latent_channels,
                    num_heads,
                    head_dim,
                    dropout=dropout,
                    context_dim=context_dim,
                    use_checkpoint=use_checkpoint,
                    attn_split_chunk=attn_split_chunk,
                )
                for _ in range(num_layers)
            ]
        )
        self.from_latent = zero_module(nn.Conv2d(latent_channels, in_channels, 1, 1, 0))

    def forward(self, net: Tensor, context: Optional[Tensor]) -> Tensor:
        inp = net
        b, c, h, w = net.shape
        net = self.norm(net)
        net = self.to_latent(net)
        net = net.permute(0, 2, 3, 1).reshape(b, h * w, c)
        for block in self.blocks:
            net = block(net, context=context)
        net = net.permute(0, 2, 1).contiguous()
        net = net.view(b, c, h, w)
        net = self.from_latent(net)
        return inp + net


__all__ = [
    "MixedStackedEncoder",
    "SpatialTransformer",
]



import os

WORKPLACE_ENVIRON_KEY = "CFLEARN_WORKPLACE"

LOSS_KEY = "loss"
INPUT_KEY = "input"
LATENT_KEY = "latent"
PREDICTIONS_KEY = "predictions"
LABEL_KEY = "labels"
ORIGINAL_LABEL_KEY = "original_labels"
BATCH_INDICES_KEY = "batch_indices"

PT_PREFIX = "model_"
SCORES_FILE = "scores.json"
CHECKPOINTS_FOLDER = "checkpoints"

META_CONFIG_NAME = "__meta__"
ML_PIPELINE_SAVE_NAME = "ml_pipeline"

DEFAULT_ZOO_TAG = "default"
CACHE_DIR = os.path.join(os.path.expanduser("~"), ".cache", "carefree-learn")
DATA_CACHE_DIR = os.path.join(CACHE_DIR, "data")

# style transfer constants

STYLE_KEY = "style"
STYLE_LATENTS_KEY = "style_latents"
CONTENT_LATENT_KEY = "content_latent"
STYLIZED_STYLE_LATENTS_KEY = "stylized_style_latents"
STYLIZED_CONTENT_LATENT_KEY = "stylized_content_latent"

INPUT_B_KEY = "input_b"
LABEL_B_KEY = "labels_b"



import os
import torch

import numpy as np

from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Callable
from typing import Optional
from typing import Protocol
from collections import OrderedDict
from cftool.misc import lock_manager
from cftool.misc import safe_execute
from cftool.misc import Saving
from cftool.array import softmax
from cftool.array import get_full_logits
from cftool.array import get_label_predictions
from cftool.types import np_dict_type

from ...data import MLData
from ...data import MLLoader
from ...data import DataModule
from ...types import data_type
from ...types import configs_type
from ...types import sample_weights_type
from ...types import states_callback_type
from ...trainer import get_sorted_checkpoints
from ...pipeline import get_requirements
from ...pipeline import IModifier
from ...pipeline import DLPipeline
from ...protocol import InferenceOutputs
from ...constants import PREDICTIONS_KEY
from ...data.ml import IMLData
from ...data.ml import MLCarefreeData
from ...data.ml import IMLDataProcessor
from ...data.ml import _InternalCarefreeMLDataProcessor
from ...misc.toolkit import ConfigMeta
from ...misc.internal_.inference import MLInference
from ...models.ml.encoders import Encoder
from ...models.ml.encoders import EncodingSettings
from ...models.protocols.ml import MLModel

try:
    from cfdata.tabular import ColumnTypes
    from cfdata.tabular import TabularData
    from cfml.misc.toolkit import ModelPattern
except:
    ColumnTypes = TabularData = ModelPattern = None


class IMLPredict(Protocol):
    def __call__(
        self,
        data: IMLData,
        *,
        return_classes: bool = False,
        binary_threshold: float = 0.5,
        return_probabilities: bool = False,
        contains_labels: bool = True,
        **predict_kwargs: Any,
    ) -> np_dict_type:
        pass


class IMLMakeModifier(Protocol):
    def __call__(self) -> "MLModifier":
        pass


class IMLMakeInferenceData(Protocol):
    def __call__(
        self,
        x: data_type,
        y: data_type = None,
        *,
        shuffle: bool = False,
        sample_weights: sample_weights_type = None,
    ) -> IMLData:
        pass


class IMLPipeline:
    processor: IMLDataProcessor

    is_classification: bool
    output_dim: Optional[int]
    use_auto_loss: bool

    encoder: Optional[Encoder]
    encoder_config: Dict[str, Any]
    encoding_settings: Dict[int, EncodingSettings]
    numerical_columns: List[int]
    categorical_columns: List[int]
    use_one_hot: bool
    use_embedding: bool
    only_categorical: bool
    use_encoder_cache: bool

    core_name: str
    core_config: Dict[str, Any]

    _pre_process_batch: bool
    _num_repeat: Optional[int]
    _make_modifier: IMLMakeModifier

    predict: IMLPredict
    make_inference_data: IMLMakeInferenceData


_ml_requirements = get_requirements(IMLPipeline, excludes=[])
_ml_build_steps = ["setup_processor", "setup_defaults", "setup_encoder"]


@IModifier.register("ml")
class MLModifier(IModifier, IMLPipeline):
    build_steps = _ml_build_steps + IModifier.build_steps
    requirements = IModifier.requirements + _ml_requirements

    # build steps

    def setup_processor(self, data_info: Dict[str, Any]) -> None:
        self.processor = data_info["processor"]

    def setup_defaults(self, data_info: Dict[str, Any]) -> None:
        self.is_classification = data_info["is_classification"]
        is_reg = not self.is_classification
        if self.input_dim is None:
            self.input_dim = data_info["input_dim"]
            self._defaults["input_dim"] = self.input_dim
        if self.output_dim is None:
            self.output_dim = 1 if is_reg else data_info["num_classes"]
            if self.output_dim is None:
                raise ValueError(
                    "either `MLData` with `cf_data`, or `output_dim`, "
                    "should be provided"
                )
            self._defaults["output_dim"] = self.output_dim
        self.use_auto_loss = self.loss_name == "auto"
        if self.use_auto_loss:
            if is_reg:
                self.loss_name = "mae"
            else:
                self.loss_name = "bce" if self.output_dim == 1 else "focal"
            self._defaults["loss_name"] = self.loss_name

    def setup_encoder(self, data_info: Dict[str, Any]) -> None:
        assert isinstance(self.input_dim, int)
        all_indices = list(range(self.input_dim))
        if not self.encoding_settings:
            self.encoder = None
            self.numerical_columns = all_indices
            self.categorical_columns = []
            self.use_one_hot = False
            self.use_embedding = False
        else:
            use_one_hot = False
            use_embedding = False
            for setting in self.encoding_settings.values():
                use_one_hot = use_one_hot or setting.use_one_hot
                use_embedding = use_embedding or setting.use_embedding
            self.encoder = self._instantiate_encoder(self.encoding_settings)
            self.use_one_hot = use_one_hot
            self.use_embedding = use_embedding
            categorical_columns = self.encoder.columns.copy()
            categorical_set = set(categorical_columns)
            numerical_columns = [i for i in all_indices if i not in categorical_set]
            self.numerical_columns = sorted(numerical_columns)
            self.categorical_columns = sorted(categorical_columns)

    def _instantiate_encoder(self, settings: Dict[int, EncodingSettings]) -> Encoder:
        if self.in_loading or not self.use_encoder_cache:
            loaders = []
        else:
            if self.data is None:
                raise ValueError("`data` should be prepared when `in_loading` is False")
            train_loader, valid_loader = self.data.initialize()
            train_loader_copy = train_loader.copy()
            train_loader_copy.disable_shuffle()
            loaders = [train_loader_copy]
            if valid_loader is not None:
                assert isinstance(valid_loader, MLLoader)
                loaders.append(valid_loader)
        return Encoder(settings, config=self.encoder_config, loaders=loaders)

    def build_model(self, data_info: Dict[str, Any]) -> None:
        assert isinstance(self.input_dim, int)
        assert isinstance(self.output_dim, int)
        self.model = MLModel(
            self.output_dim,
            data_info["num_history"],
            encoder=self.encoder,
            use_encoder_cache=self.use_encoder_cache,
            numerical_columns=self.numerical_columns,
            categorical_columns=self.categorical_columns,
            use_one_hot=self.use_one_hot,
            use_embedding=self.use_embedding,
            only_categorical=self.only_categorical,
            core_name=self.core_name,
            core_config=self.core_config,
            pre_process_batch=self._pre_process_batch,
            num_repeat=self._num_repeat,
        )

    def build_inference(self) -> None:
        self.inference = MLInference(model=self.model)

    def prepare_trainer_defaults(self, data_info: Dict[str, Any]) -> None:  # type: ignore
        if self.trainer_config["monitor_names"] is None:
            self.trainer_config["monitor_names"] = ["mean_std", "plateau"]
            self._defaults["monitor_names"] = ["mean_std", "plateau"]
        super().prepare_trainer_defaults(data_info)
        if (
            self.trainer_config["metric_names"] is None
            and self.use_auto_loss
            and not self.trainer_config["use_losses_as_metrics"]
        ):
            if self.is_classification:
                self.trainer_config["metric_names"] = ["acc", "auc"]
            else:
                self.trainer_config["metric_names"] = ["mae", "mse"]
            self._defaults["metric_names"] = self.trainer_config["metric_names"]
        callback_names = self.trainer_config["callback_names"]
        if "_inject_loader_name" not in callback_names:
            callback_names.append("_inject_loader_name")
            cbs = self._defaults.setdefault("additional_callbacks", [])
            cbs.append("_inject_loader_name")

    ## api

    def build(
        self,
        data_info: Dict[str, Any],
        *,
        build_trainer: bool = True,
        report: Optional[bool] = None,
    ) -> None:
        super().build(data_info, build_trainer=build_trainer, report=report)
        self.processor.is_ready = True

    # load steps

    def permute_states(self, states: Dict[str, Any]) -> None:
        if self.encoder is not None:
            encoder_cache_keys = []
            for key in states:
                if key.startswith("encoder") and key.endswith("cache"):
                    encoder_cache_keys.append(key)
            for key in encoder_cache_keys:
                states.pop(key)

    # inference

    def postprocess(  # type: ignore
        self,
        outputs: InferenceOutputs,
        *,
        return_classes: bool = False,
        binary_threshold: float = 0.5,
        return_probabilities: bool = False,
    ) -> np_dict_type:
        forward = super().postprocess(outputs)
        if not self.is_classification:
            pass
        elif return_classes and return_probabilities:
            raise ValueError(
                "`return_classes` & `return_probabilities`"
                "should not be True at the same time"
            )
        elif not return_classes and not return_probabilities:
            pass
        else:
            predictions = forward[PREDICTIONS_KEY]
            if predictions.shape[1] > 2 and return_classes:
                forward[PREDICTIONS_KEY] = predictions.argmax(1, keepdims=True)
            else:
                probabilities = softmax(predictions)
                if return_probabilities:
                    forward[PREDICTIONS_KEY] = probabilities
                else:
                    assert probabilities.shape[1] == 2, "internal error occurred"
                    classes = (probabilities[..., [1]] >= binary_threshold).astype(int)
                    forward[PREDICTIONS_KEY] = classes
        # processor callback
        kw = dict(
            forward=forward,
            return_classes=return_classes,
            binary_threshold=binary_threshold,
            return_probabilities=return_probabilities,
        )
        forward = safe_execute(self.processor.postprocess_results, kw)
        return forward


@DLPipeline.register("ml")
class MLPipeline(IMLPipeline, DLPipeline, metaclass=ConfigMeta):  # type: ignore
    modifier = "ml"

    data: MLData
    model: MLModel
    inference: MLInference
    inference_base = MLInference

    def __init__(
        self,
        core_name: str = "fcnn",
        core_config: Optional[Dict[str, Any]] = None,
        *,
        input_dim: Optional[int] = None,
        output_dim: Optional[int] = None,
        loss_name: str = "auto",
        loss_config: Optional[Dict[str, Any]] = None,
        # encoder
        use_encoder_cache: bool = False,
        only_categorical: bool = False,
        encoder_config: Optional[Dict[str, Any]] = None,
        encoding_settings: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        # trainer
        state_config: Optional[Dict[str, Any]] = None,
        num_epoch: int = 40,
        max_epoch: int = 1000,
        fixed_epoch: Optional[int] = None,
        fixed_steps: Optional[int] = None,
        log_steps: Optional[int] = None,
        valid_portion: float = 1.0,
        amp: bool = False,
        clip_norm: float = 0.0,
        cudnn_benchmark: bool = False,
        metric_names: Optional[Union[str, List[str]]] = None,
        metric_configs: configs_type = None,
        metric_weights: Optional[Dict[str, float]] = None,
        use_losses_as_metrics: Optional[bool] = None,
        loss_metrics_weights: Optional[Dict[str, float]] = None,
        recompute_train_losses_in_eval: bool = True,
        monitor_names: Optional[Union[str, List[str]]] = None,
        monitor_configs: Optional[Dict[str, Any]] = None,
        callback_names: Optional[Union[str, List[str]]] = None,
        callback_configs: Optional[Dict[str, Any]] = None,
        lr: Optional[float] = None,
        optimizer_name: Optional[str] = None,
        scheduler_name: Optional[str] = None,
        optimizer_config: Optional[Dict[str, Any]] = None,
        scheduler_config: Optional[Dict[str, Any]] = None,
        update_scheduler_per_epoch: bool = False,
        optimizer_settings: Optional[Dict[str, Dict[str, Any]]] = None,
        use_zero: bool = False,
        workplace: str = "_logs",
        finetune_config: Optional[Dict[str, Any]] = None,
        tqdm_settings: Optional[Dict[str, Any]] = None,
        # misc
        in_loading: bool = False,
        pre_process_batch: bool = True,
        num_repeat: Optional[int] = None,
    ):
        super().__init__(
            "MLModel",
            loss_name=loss_name,
            loss_config=loss_config,
            state_config=state_config,
            num_epoch=num_epoch,
            max_epoch=max_epoch,
            fixed_epoch=fixed_epoch,
            fixed_steps=fixed_steps,
            log_steps=log_steps,
            valid_portion=valid_portion,
            amp=amp,
            clip_norm=clip_norm,
            cudnn_benchmark=cudnn_benchmark,
            metric_names=metric_names,
            metric_configs=metric_configs,
            metric_weights=metric_weights,
            use_losses_as_metrics=use_losses_as_metrics,
            loss_metrics_weights=loss_metrics_weights,
            recompute_train_losses_in_eval=recompute_train_losses_in_eval,
            monitor_names=monitor_names,
            monitor_configs=monitor_configs,
            callback_names=callback_names,
            callback_configs=callback_configs,
            lr=lr,
            optimizer_name=optimizer_name,
            scheduler_name=scheduler_name,
            optimizer_config=optimizer_config,
            scheduler_config=scheduler_config,
            update_scheduler_per_epoch=update_scheduler_per_epoch,
            optimizer_settings=optimizer_settings,
            use_zero=use_zero,
            workplace=workplace,
            finetune_config=finetune_config,
            tqdm_settings=tqdm_settings,
            in_loading=in_loading,
        )
        self.core_name = core_name
        self.core_config = core_config or {}
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.use_encoder_cache = use_encoder_cache
        self.only_categorical = only_categorical
        self.encoder = None
        self.encoder_config = encoder_config or {}
        self.encoding_settings = {}
        for idx, raw_setting in (encoding_settings or {}).items():
            # when loading from json, the keys (idx) will be str
            self.encoding_settings[int(idx)] = EncodingSettings(**raw_setting)
        self._pre_process_batch = pre_process_batch
        self._num_repeat = num_repeat

    def make_inference_data(  # type: ignore
        self,
        x: data_type,
        y: data_type = None,
        *,
        shuffle: bool = False,
        sample_weights: sample_weights_type = None,
        **kwargs: Any,
    ) -> IMLData:
        kwargs["processor"] = self.data_info["processor"]
        kwargs["shuffle_train"] = shuffle
        kwargs["for_inference"] = True
        data: IMLData = IMLData.get(self.data_info["type"])(x, y, **kwargs)
        data.prepare(sample_weights)
        return data

    def to_pattern(
        self,
        *,
        binary_threshold: float = 0.5,
        pre_process: Optional[Callable[[IMLData], IMLData]] = None,
        **predict_kwargs: Any,
    ) -> ModelPattern:
        if ModelPattern is None:
            raise ValueError("`carefree-ml` need to be installed to use `to_pattern`")

        def _predict(idata: IMLData) -> np.ndarray:
            if pre_process is not None:
                idata = pre_process(idata)
            predictions = self.predict(idata, **predict_kwargs)[PREDICTIONS_KEY]
            if self.is_classification:
                return get_label_predictions(predictions, binary_threshold)
            return predictions

        def _predict_prob(idata: IMLData) -> np.ndarray:
            if not self.is_classification:
                msg = "`predict_prob` should not be called in regression tasks"
                raise ValueError(msg)
            if pre_process is not None:
                idata = pre_process(idata)
            logits = self.predict(idata, **predict_kwargs)[PREDICTIONS_KEY]
            logits = get_full_logits(logits)
            return softmax(logits)

        return ModelPattern(
            init_method=lambda: self,
            predict_method=_predict,
            predict_prob_method=_predict_prob,
        )

    @staticmethod
    def fuse_multiple(
        export_folders: List[str],
        *,
        cuda: Optional[str] = None,
        compress: bool = True,
        build_trainer: bool = False,
        report: Optional[bool] = None,
        states_callback: states_callback_type = None,
        pre_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        post_callback: Optional[Callable[[DLPipeline, Dict[str, Any]], None]] = None,
    ) -> "MLPipeline":
        export_folder = export_folders[0]
        base_folder = os.path.dirname(os.path.abspath(export_folder))
        with lock_manager(base_folder, [export_folder]):
            with Saving.compress_loader(export_folder, compress):
                m = IModifier.load_infrastructure(
                    MLPipeline,
                    export_folder,
                    cuda,
                    False,
                    pre_callback,
                    post_callback,
                )
                loaded = DataModule.load_info(export_folder, return_bytes=True)
                data_info = loaded.info
                m.data_module_bytes = loaded.data_module_bytes
        m._num_repeat = m.config["num_repeat"] = len(export_folders)
        m._make_modifier().build(data_info, build_trainer=build_trainer, report=report)
        m.model.to(m.device)
        merged_states: OrderedDict[str, torch.Tensor] = OrderedDict()
        for i, export_folder in enumerate(export_folders):
            base_folder = os.path.dirname(os.path.abspath(export_folder))
            with lock_manager(base_folder, [export_folder]):
                with Saving.compress_loader(export_folder, compress):
                    checkpoints = get_sorted_checkpoints(export_folder)
                    checkpoint_path = os.path.join(export_folder, checkpoints[0])
                    states = torch.load(checkpoint_path, map_location=m.device)
                    current_keys = list(states.keys())
                    for k, v in list(states.items()):
                        split = k.split(".")
                        split.insert(1, str(i))
                        states[".".join(split)] = v
                    for k in current_keys:
                        states.pop(k)
                    if states_callback is not None:
                        states = states_callback(m, states)
                    merged_states.update(states)
        m.model.load_state_dict(merged_states)
        return m


# carefree


class IMLCarefreeMakeModifier(Protocol):
    def __call__(self) -> "MLCarefreeModifier":
        pass


class IMLCarefreeMakeInferenceData(Protocol):
    def __call__(
        self,
        x: data_type,
        y: data_type = None,
        *,
        shuffle: bool = False,
        sample_weights: sample_weights_type = None,
        contains_labels: bool = True,
    ) -> MLCarefreeData:
        pass


class IMLCarefreePipeline:
    processor: _InternalCarefreeMLDataProcessor

    _make_modifier: IMLCarefreeMakeModifier
    make_inference_data: IMLCarefreeMakeInferenceData


_ml_carefree_requirements = get_requirements(IMLCarefreePipeline, excludes=[])


@IModifier.register("ml.carefree")
class MLCarefreeModifier(IMLCarefreePipeline, MLModifier):  # type: ignore
    requirements = MLModifier.requirements + _ml_carefree_requirements

    # build steps

    def setup_encoder(self) -> None:  # type: ignore
        excluded = 0
        encoding_settings = {}
        numerical_columns = []
        categorical_columns = []
        use_one_hot = False
        use_embedding = False
        cf_data = self.processor.cf_data
        if cf_data.is_simplify:
            numerical_columns = list(range(cf_data.processed.x.shape[1]))
        else:
            ts_indices = cf_data.ts_indices
            recognizers = cf_data.recognizers
            sorted_indices = [idx for idx in sorted(recognizers) if idx != -1]
            for idx in sorted_indices:
                recognizer = recognizers[idx]
                assert recognizer is not None
                if not recognizer.info.is_valid or idx in ts_indices:
                    excluded += 1
                elif recognizer.info.column_type is ColumnTypes.NUMERICAL:
                    numerical_columns.append(idx - excluded)
                elif idx not in encoding_settings:
                    true_idx = idx - excluded
                    setting = EncodingSettings(dim=recognizer.num_unique_values)
                    encoding_settings[true_idx] = setting
                    use_one_hot = use_one_hot or setting.use_one_hot
                    use_embedding = use_embedding or setting.use_embedding
                    categorical_columns.append(true_idx)
        if not encoding_settings:
            encoder = None
        else:
            encoder = self._instantiate_encoder(encoding_settings)
        self.encoder = encoder
        self.use_one_hot = use_one_hot
        self.use_embedding = use_embedding
        self.numerical_columns = numerical_columns
        self.categorical_columns = categorical_columns


@DLPipeline.register("ml.carefree")
class MLCarefreePipeline(IMLCarefreePipeline, MLPipeline):  # type: ignore
    modifier = "ml.carefree"

    @property
    def cf_data(self) -> TabularData:
        return self.processor.cf_data


__all__ = [
    "MLModifier",
    "MLPipeline",
    "MLCarefreeModifier",
    "MLCarefreePipeline",
]



import torch

import torch.nn as nn

from typing import Any
from typing import Dict
from typing import Optional

from ...protocols.cv import ImageTranslatorMixin
from ....modules.blocks import NormFactory
from ....misc.internal_.register import register_module


class UnetSkipConnectionBlock(nn.Module):
    """
    Dimension flows:
           input                   (input_channels)
        -> down_conv
        -> inner_net               (inner_channels)
        -> submodule
        -> sub_out                 (submodule.outer_channels)
        -> cat(inner_net, sub_out) (submodule.input_channels + submodule.outer_channels)
        -> up_conv
        -> output                  (outer_channels)

    > This means that:
    1) inner_channels should be submodule.input_channels
    2) we should restrict submodule.input_channels & submodule.outer_channels
    -> e.g. let submodule.input_channels = submodule.outer_channels = inner_channels
    """

    def __init__(
        self,
        inner_channels: int,
        outer_channels: int,
        input_channels: Optional[int] = None,
        submodule: Optional["UnetSkipConnectionBlock"] = None,
        is_outermost: bool = False,
        is_innermost: bool = False,
        norm_type: Optional[str] = "batch",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        use_dropout: bool = False,
    ):
        super().__init__()
        self.inner_channels = inner_channels
        self.outer_channels = outer_channels
        self.input_channels = input_channels or outer_channels
        self.is_outermost = is_outermost
        msg = None
        if submodule:
            if submodule.is_outermost:
                msg = "submodule is already outermost"
            elif inner_channels != submodule.input_channels:
                msg = "`inner_channels` should equal to `input_channels` of submodule"
            elif inner_channels != submodule.outer_channels:
                msg = "`inner_channels` should equal to `outer_channels` of submodule"
        if msg is not None:
            raise ValueError(msg)
        use_bias = norm_type == "instance"
        norm_kwargs = norm_kwargs or {}
        downsample = nn.Conv2d(
            self.input_channels,
            inner_channels,
            kernel_size=4,
            stride=2,
            padding=1,
            bias=use_bias,
        )
        norm_factory = NormFactory(norm_type)
        if is_outermost:
            blocks = [
                downsample,
                submodule,
                nn.ReLU(True),
                nn.ConvTranspose2d(
                    inner_channels * 2,
                    outer_channels,
                    kernel_size=4,
                    stride=2,
                    padding=1,
                ),
                nn.Tanh(),
            ]
        elif is_innermost:
            blocks = [
                nn.LeakyReLU(0.2, True),
                downsample,
                nn.ReLU(True),
                nn.ConvTranspose2d(
                    inner_channels,
                    outer_channels,
                    kernel_size=4,
                    stride=2,
                    padding=1,
                    bias=use_bias,
                ),
                norm_factory.make(outer_channels, **norm_kwargs),
            ]
        else:
            blocks = [
                nn.LeakyReLU(0.2, True),
                downsample,
                norm_factory.make(inner_channels, **norm_kwargs),
                submodule,
                nn.ReLU(True),
                nn.ConvTranspose2d(
                    inner_channels * 2,
                    outer_channels,
                    kernel_size=4,
                    stride=2,
                    padding=1,
                    bias=use_bias,
                ),
                norm_factory.make(outer_channels, **norm_kwargs),
            ]
            if use_dropout:
                blocks.append(nn.Dropout(0.5))

        self.model = nn.Sequential(*blocks)

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        if self.is_outermost:
            return self.model(net)
        return torch.cat([net, self.model(net)], 1)


@register_module("unet_generator")
class UnetGenerator(nn.Module, ImageTranslatorMixin):
    def __init__(
        self,
        in_channels: int = 3,
        out_channels: Optional[int] = None,
        num_downsample: int = 8,
        *,
        start_channels: int = 64,
        norm_type: Optional[str] = "batch",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        use_dropout: bool = True,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels or in_channels
        self.num_downsample = num_downsample
        self.start_channels = start_channels

        unet_block = UnetSkipConnectionBlock(
            start_channels * 8,
            start_channels * 8,
            input_channels=None,
            submodule=None,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
            is_innermost=True,
        )
        for i in range(num_downsample - 5):
            unet_block = UnetSkipConnectionBlock(
                start_channels * 8,
                start_channels * 8,
                input_channels=None,
                submodule=unet_block,
                norm_type=norm_type,
                norm_kwargs=norm_kwargs,
                use_dropout=use_dropout,
            )
        unet_block = UnetSkipConnectionBlock(
            start_channels * 8,
            start_channels * 4,
            input_channels=None,
            submodule=unet_block,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
        )
        unet_block = UnetSkipConnectionBlock(
            start_channels * 4,
            start_channels * 2,
            input_channels=None,
            submodule=unet_block,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
        )
        unet_block = UnetSkipConnectionBlock(
            start_channels * 2,
            start_channels,
            input_channels=None,
            submodule=unet_block,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
        )
        self.model = UnetSkipConnectionBlock(
            start_channels,
            self.out_channels,
            input_channels=in_channels,
            submodule=unet_block,
            is_outermost=True,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
        )

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        return self.model(net)


__all__ = ["UnetGenerator"]



import torch

import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Dict
from typing import List
from typing import Tuple
from typing import Optional
from itertools import combinations

from ..register import register_ml_module
from ...misc.toolkit import eval_context
from ...modules.blocks import Activation


class SeparableMapping(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        num_groups: int,
        *,
        bias: bool = True,
        activation: Optional[str] = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.num_groups = num_groups
        in_dim *= num_groups
        out_dim *= num_groups
        self.conv = nn.Conv1d(in_dim, out_dim, 1, groups=num_groups, bias=bias)
        self.bn = None if not batch_norm else nn.BatchNorm1d(out_dim)
        self.activation = Activation.make(activation)
        use_dropout = 0.0 < dropout < 1.0
        self.dropout = None if not use_dropout else nn.Dropout(dropout)

    # [B, num_groups * in_dim, D] -> [B, num_groups * out_dim, D]
    # note that they should be originated from [B, num_groups, in_dim / out_dim, D]
    def forward(self, net: Tensor) -> Tensor:
        net = self.conv(net)
        if self.bn is not None:
            net = self.bn(net)
        net = self.activation(net)
        if self.dropout is not None:
            net = self.dropout(net)
        return net


class NBMBlock(nn.Module):
    def __init__(
        self,
        order: int,
        num_bases: int,
        *,
        hidden_units: List[int],
        bias: bool = True,
        activation: str = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
    ) -> None:
        super().__init__()
        self.order = order
        self.num_bases = num_bases

        blocks: List[nn.Module] = []
        in_dim = order
        for hidden_unit in hidden_units:
            mapping = SeparableMapping(
                in_dim,
                hidden_unit,
                num_bases,
                bias=bias,
                activation=activation,
                batch_norm=batch_norm,
                dropout=dropout,
            )
            blocks.append(mapping)
            in_dim = hidden_unit
        blocks.append(
            SeparableMapping(
                in_dim,
                1,
                num_bases,
                bias=bias,
                activation=None,
                batch_norm=False,
                dropout=0.0,
            )
        )

        self.net = nn.Sequential(*blocks)

    def forward(self, net: Tensor, indices: List[Tuple[int, ...]]) -> Tensor:
        for i, pack in enumerate(indices):
            if len(pack) != self.order:
                raise ValueError(
                    f"length of every tuple in indices must be {self.order}, "
                    f"but got {pack} at index {i}"
                )
        B = len(net)
        D = len(indices)
        # [B, D, order]
        net = net[..., indices]
        # [B, order, D]
        net = net.transpose(1, 2)
        # [B, 1, order, D]
        net = net.reshape(B, 1, self.order, D)
        # [B, num_bases, order, D]
        net = net.repeat_interleave(self.num_bases, 1)
        # [B, num_bases * order, D]
        net = net.contiguous().view(B, self.num_bases * self.order, D)
        # [B, num_bases, D]
        net = self.net(net)
        # [B, D, num_bases]
        return net.transpose(1, 2).contiguous()


@register_ml_module("nbm")
class NBM(nn.Module):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int,
        *,
        nary: Optional[Dict[str, List[Tuple[int, ...]]]] = None,
        num_bases: int = 64,
        bases_dropout: float = 0.0,
        hidden_units: Optional[List[int]] = None,
        bias: bool = True,
        activation: str = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
    ):
        super().__init__()
        input_dim *= num_history
        if hidden_units is None:
            dim = max(32, min(1024, 2 * input_dim))
            hidden_units = 2 * [dim]

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.bias = bias

        self.bases = nn.ModuleDict()
        self.nary = nary or {"1": list(combinations(range(input_dim), 1))}
        self.num_bases = num_bases
        for order in self.nary:
            self.bases[order] = NBMBlock(
                int(order),
                num_bases,
                hidden_units=hidden_units,
                bias=bias,
                activation=activation,
                batch_norm=batch_norm,
                dropout=dropout,
            )
        self.bases_dropout = nn.Dropout(p=bases_dropout)

        self.all_indices = []
        for indices in self.nary.values():
            self.all_indices += indices
        self.latent_dim = len(self.all_indices)

        self.featurizer = nn.Conv1d(
            self.latent_dim * num_bases,
            self.latent_dim,
            kernel_size=1,
            groups=self.latent_dim,
            bias=bias,
        )
        self.head = nn.Linear(self.latent_dim, output_dim, bias=True)

    def get_key(self, order: str) -> str:
        return f"order_{order}"

    def forward(self, net: Tensor, *, return_features: bool = False) -> Tensor:
        B = len(net)
        if len(net.shape) > 2:
            net = net.contiguous().view(B, -1)
        bases = []
        for order, indices in self.nary.items():
            base = self.bases[order](net, indices)
            base = self.bases_dropout(base)
            bases.append(base)
        # [B, latent_dim, num_bases]
        net = torch.cat(bases, dim=1)
        # [B, latent_dim * num_bases, 1]
        net = net.view(B, self.latent_dim * self.num_bases, 1)
        # [B, latent_dim]
        net = self.featurizer(net).squeeze(2)
        if return_features:
            return net
        # [B, out_dim]
        net = self.head(net)
        return net

    def inspect(
        self,
        net: Tensor,
        x_dims: Tuple[int, ...],
        y_dim: int,
        *,
        already_extracted: bool = False,
    ) -> Tensor:
        try:
            idx = self.all_indices.index(x_dims)
        except ValueError:
            raise ValueError(f"`x_dims` ({x_dims}) is not in the pre-defined nary")
        with eval_context(self):
            B = len(net)
            if len(net.shape) > 2:
                net = net.contiguous().view(B, -1)
            order = len(x_dims)
            # [B, 1, num_bases]
            indices = [tuple(range(order))] if already_extracted else [x_dims]
            net = self.bases[str(order)](net, indices)
            net = self.bases_dropout(net)
            # [B, num_bases]
            net = net.view(B, self.num_bases, 1).squeeze(2)
            # [B, 1]
            w = self.featurizer.weight[[idx]].squeeze(2)
            b = None if not self.bias else self.featurizer.bias[[idx]]
            net = F.linear(net, w, b)
            # [B, 1]
            w = self.head.weight[..., [idx]][[y_dim]]
            net = F.linear(net, w)
        return net


__all__ = ["NBM"]



from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from cftool.types import np_dict_type

from .pipeline import CVPipeline
from ...data import Transforms
from ...data import InferenceImagePathsData
from ...data import InferenceImageFolderData


class InferenceResults(NamedTuple):
    outputs: np_dict_type
    img_paths: List[str]


def predict_paths(
    m: CVPipeline,
    paths: List[str],
    *,
    batch_size: int,
    num_workers: int = 0,
    transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
    transform_config: Optional[Dict[str, Any]] = None,
    use_tqdm: bool = True,
    **predict_kwargs: Any,
) -> InferenceResults:
    idata = InferenceImagePathsData(
        paths,
        batch_size=batch_size,
        num_workers=num_workers,
        transform=m.test_transform or transform,
        transform_config=transform_config,
    )
    outputs = m.predict(idata, use_tqdm=use_tqdm, **predict_kwargs)
    return InferenceResults(outputs, idata.dataset.img_paths)


def predict_folder(
    m: CVPipeline,
    folder: str,
    *,
    batch_size: int,
    num_workers: int = 0,
    transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
    transform_config: Optional[Dict[str, Any]] = None,
    use_tqdm: bool = True,
    **predict_kwargs: Any,
) -> InferenceResults:
    idata = InferenceImageFolderData(
        folder,
        batch_size=batch_size,
        num_workers=num_workers,
        transform=m.test_transform or transform,
        transform_config=transform_config,
    )
    outputs = m.predict(idata, use_tqdm=use_tqdm, **predict_kwargs)
    return InferenceResults(outputs, idata.dataset.img_paths)


__all__ = [
    "predict_paths",
    "predict_folder",
]



import torch

import numpy as np
import torch.nn.functional as F

from PIL import Image
from torch import Tensor
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from cftool.array import save_images

from .common import restrict_wh
from .common import APIMixin
from ....pipeline import DLPipeline
from ....zoo.core import DLZoo
from ....misc.toolkit import eval_context

try:
    from cfcv.misc.toolkit import to_rgb
except:
    to_rgb = None


def default_preprocess(image: Image.Image) -> np.ndarray:
    return np.array(image).astype(np.float32) / 255.0


class TranslatorAPI(APIMixin):
    def sr(
        self,
        image: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        max_wh: int = 768,
        clip_range: Optional[Tuple[int, int]] = (0, 1),
        preprocess_fn: Optional[Callable[[Image.Image], np.ndarray]] = None,
    ) -> Tensor:
        if isinstance(image, str):
            image = Image.open(image)
        w, h = image.size
        image = image.resize(restrict_wh(w, h, max_wh), resample=Image.LANCZOS)
        # handle alpha
        alpha = None
        if image.mode == "RGBA":
            if to_rgb is None:
                raise ValueError("`carefree-cv` is needed for `TranslatorAPI`")
            alpha = image.split()[3]
            image = to_rgb(image)
        # inference
        array = (preprocess_fn or default_preprocess)(image)
        tensor = torch.from_numpy(array)[None].permute(0, 3, 1, 2)
        tensor = tensor.contiguous().to(self.device)
        if self.use_half:
            tensor = tensor.half()
        with eval_context(self.m):
            output = self.m(tensor).cpu().float()
        if clip_range is not None:
            output = torch.clip(output, *clip_range)
        # handle alpha
        if alpha is not None:
            alpha_tensor = torch.from_numpy(np.array(alpha).astype(np.float32) / 255.0)
            alpha_tensor = alpha_tensor[None, None]
            with torch.no_grad():
                alpha_tensor = F.interpolate(
                    alpha_tensor,
                    output.shape[-2:],
                    mode="nearest",
                )
            output = torch.cat([output, alpha_tensor], dim=1)
        # export
        if export_path is not None:
            save_images(output, export_path)
        torch.cuda.empty_cache()
        return output

    @classmethod
    def from_esr(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "TranslatorAPI":
        return cls.from_pipeline(esr(), device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_esr_anime(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "TranslatorAPI":
        m = esr_anime()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)


def esr(pretrained: bool = True) -> DLPipeline:
    return DLZoo.load_pipeline("sr/esr", pretrained=pretrained)


def esr_anime(pretrained: bool = True) -> DLPipeline:
    return DLZoo.load_pipeline("sr/esr.anime", pretrained=pretrained)


__all__ = [
    "TranslatorAPI",
]



import math
import torch

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional

from .utils import append_dims
from .protocol import ISampler
from .protocol import IQSampler
from .protocol import IDiffusion
from .protocol import UncondSamplerMixin


@ISampler.register("solver")
class DPMSolver(ISampler, UncondSamplerMixin):
    def __init__(
        self,
        model: IDiffusion,
        *,
        schedule: str = "linear",
        unconditional_cond: Optional[Any] = None,
        unconditional_guidance_scale: float = 1.0,
        t0: float = 1.0e-4,
        tT: Optional[float] = None,
        order: int = 3,
        skip_type: str = "logSNR",
        fast_version: bool = True,
        default_steps: int = 20,
    ):
        if model.parameterization != "eps":
            raise ValueError("only `eps` parameterization is supported in `ddim`")
        super().__init__(model)
        if schedule not in ["linear", "cosine"]:
            raise ValueError("only `linear` & `cosine` can be used as `schedule`")
        self.unconditional_cond = unconditional_cond
        self.unconditional_guidance_scale = unconditional_guidance_scale
        self.beta_0 = 0.1
        self.beta_1 = 20
        self.cosine_s = 0.008
        self.cosine_beta_max = 999.0
        self.cosine_t_max = (
            math.atan(self.cosine_beta_max * (1.0 + self.cosine_s) / math.pi)
            * 2.0
            * (1.0 + self.cosine_s)
            / math.pi
            - self.cosine_s
        )
        self.cosine_log_alpha_0 = math.log(
            math.cos(self.cosine_s / (1.0 + self.cosine_s) * math.pi / 2.0)
        )
        self.default_tT = 0.9946 if schedule == "cosine" else 1.0
        self.schedule = schedule
        self.t0 = t0
        self.tT = tT
        self.order = order
        self.skip_type = skip_type
        self.fast_version = fast_version
        self.default_steps = default_steps

    @property
    def q_sampler(self) -> IQSampler:
        raise NotImplementedError("`DPMSolver` has not implemented `q_sampler` yet")

    @property
    def sample_kwargs(self) -> Dict[str, Any]:
        return dict(
            unconditional_cond=self.unconditional_cond,
            unconditional_guidance_scale=self.unconditional_guidance_scale,
            t0=self.t0,
            tT=self.tT,
            order=self.order,
            skip_type=self.skip_type,
            fast_version=self.fast_version,
        )

    def sample_step(
        self,
        image: Tensor,
        cond: Optional[Tensor],
        step: int,
        total_step: int,
        *,
        unconditional_cond: Optional[Any] = None,
        unconditional_guidance_scale: float = 1.0,
        t0: float = 1.0e-4,
        tT: Optional[float] = None,
        order: int = 3,
        skip_type: str = "logSNR",
        fast_version: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if tT is None:
            tT = self.default_tT
        if step == 0:
            self._reset_buffers(
                total_step,
                unconditional_cond,
                unconditional_guidance_scale,
                t0,
                tT,
                order,
                skip_type,
                fast_version,
            )
        if step >= len(self.orders):
            return image
        b = image.shape[0]
        device = image.device
        order = self.orders[step]
        vec_s = torch.ones(b, device=device) * self.timesteps[step]
        vec_t = torch.ones(b, device=device) * self.timesteps[step + 1]
        if order == 1:
            return self._order1_update(image, cond, vec_s, vec_t)
        if order == 2:
            return self._order2_update(image, cond, vec_s, vec_t)
        if order == 3:
            return self._order3_update(image, cond, vec_s, vec_t)
        raise ValueError(f"unrecognized order '{order}' occurred")

    # internal

    def _denoise(self, x: Tensor, ts: Tensor, cond: Optional[Tensor]) -> Tensor:
        ts = self.model.t * torch.max(ts - 1.0 / self.model.t, torch.zeros_like(ts))
        return self._uncond_denoise(x, ts, cond)

    def _order1_update(
        self,
        image: Tensor,
        cond: Optional[Tensor],
        vec_s: Tensor,
        vec_t: Tensor,
    ) -> Tensor:
        dims = len(image.shape) - 1
        lambda_s = self._marginal_lambda(vec_s)
        lambda_t = self._marginal_lambda(vec_t)
        h = lambda_t - lambda_s
        log_alpha_s = self._marginal_log_mean_coef(vec_s)
        log_alpha_t = self._marginal_log_mean_coef(vec_t)
        sigma_t = self._marginal_std(vec_t)
        phi_1 = torch.expm1(h)
        noise_s = self._denoise(image, vec_s, cond)
        return (
            append_dims(torch.exp(log_alpha_t - log_alpha_s), dims) * image
            - append_dims(sigma_t * phi_1, dims) * noise_s
        )

    def _order2_update(
        self,
        image: Tensor,
        cond: Optional[Tensor],
        vec_s: Tensor,
        vec_t: Tensor,
    ) -> Tensor:
        r1 = 0.5
        dims = len(image.shape) - 1
        lambda_s = self._marginal_lambda(vec_s)
        lambda_t = self._marginal_lambda(vec_t)
        h = lambda_t - lambda_s
        lambda_s1 = lambda_s + r1 * h
        s1 = self._inverse_lambda(lambda_s1)
        log_alpha_s = self._marginal_log_mean_coef(vec_s)
        log_alpha_s1 = self._marginal_log_mean_coef(s1)
        log_alpha_t = self._marginal_log_mean_coef(vec_t)
        sigma_s1 = self._marginal_std(s1)
        sigma_t = self._marginal_std(vec_t)

        phi_11 = torch.expm1(r1 * h)
        phi_1 = torch.expm1(h)

        noise_s = self._denoise(image, vec_s, cond)
        x_s1 = (
            append_dims(torch.exp(log_alpha_s1 - log_alpha_s), dims) * image
            - append_dims(sigma_s1 * phi_11, dims) * noise_s
        )
        noise_s1 = self._denoise(x_s1, s1, cond)
        return (
            append_dims(torch.exp(log_alpha_t - log_alpha_s), dims) * image
            - append_dims(sigma_t * phi_1, dims) * noise_s
            - (0.5 / r1) * append_dims(sigma_t * phi_1, dims) * (noise_s1 - noise_s)
        )

    def _order3_update(
        self,
        image: Tensor,
        cond: Optional[Tensor],
        vec_s: Tensor,
        vec_t: Tensor,
    ) -> Tensor:
        r1 = 1.0 / 3.0
        r2 = 2.0 / 3.0
        dims = len(image.shape) - 1
        lambda_s = self._marginal_lambda(vec_s)
        lambda_t = self._marginal_lambda(vec_t)
        h = lambda_t - lambda_s
        lambda_s1 = lambda_s + r1 * h
        lambda_s2 = lambda_s + r2 * h
        s1 = self._inverse_lambda(lambda_s1)
        s2 = self._inverse_lambda(lambda_s2)
        log_alpha_s = self._marginal_log_mean_coef(vec_s)
        log_alpha_s1 = self._marginal_log_mean_coef(s1)
        log_alpha_s2 = self._marginal_log_mean_coef(s2)
        log_alpha_t = self._marginal_log_mean_coef(vec_t)
        sigma_s1 = self._marginal_std(s1)
        sigma_s2 = self._marginal_std(s2)
        sigma_t = self._marginal_std(vec_t)

        phi_11 = torch.expm1(r1 * h)
        phi_12 = torch.expm1(r2 * h)
        phi_1 = torch.expm1(h)
        phi_22 = torch.expm1(r2 * h) / (r2 * h) - 1.0
        phi_2 = torch.expm1(h) / h - 1.0

        noise_s = self._denoise(image, vec_s, cond)
        x_s1 = (
            append_dims(torch.exp(log_alpha_s1 - log_alpha_s), dims) * image
            - append_dims(sigma_s1 * phi_11, dims) * noise_s
        )
        noise_s1 = self._denoise(x_s1, s1, cond)
        x_s2 = (
            append_dims(torch.exp(log_alpha_s2 - log_alpha_s), dims) * image
            - append_dims(sigma_s2 * phi_12, dims) * noise_s
            - r2 / r1 * append_dims(sigma_s2 * phi_22, dims) * (noise_s1 - noise_s)
        )
        noise_s2 = self._denoise(x_s2, s2, cond)
        return (
            append_dims(torch.exp(log_alpha_t - log_alpha_s), dims) * image
            - append_dims(sigma_t * phi_1, dims) * noise_s
            - (1.0 / r2) * append_dims(sigma_t * phi_2, dims) * (noise_s2 - noise_s)
        )

    def _reset_buffers(
        self,
        total_step: int,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        t0: float,
        tT: float,
        order: int,
        skip_type: str,
        fast_version: bool,
    ) -> None:
        if fast_version:
            K = total_step // 3 + 1
            if total_step % 3 == 0:
                orders = [3] * (K - 2) + [2, 1]
            elif total_step % 3 == 1:
                orders = [3] * (K - 1) + [1]
            else:
                orders = [3] * (K - 1) + [2]
            timesteps = self._get_time_steps("logSNR", t0, tT, K)
            self.orders = orders
            self.timesteps = timesteps
        else:
            N_steps = total_step // order
            self.orders = [order] * N_steps
            self.timesteps = self._get_time_steps(skip_type, t0, tT, N_steps)
        # unconditional conditioning
        self._reset_uncond_buffers(unconditional_cond, unconditional_guidance_scale)

    def _get_time_steps(self, skip_type: str, t0: float, tT: float, N: int) -> Tensor:
        device = self.model.device
        if skip_type == "logSNR":
            lambda_T = self._marginal_lambda(torch.tensor(tT, device=device))
            lambda_0 = self._marginal_lambda(torch.tensor(t0, device=device))
            logSNR_steps = torch.linspace(lambda_T, lambda_0, N + 1, device=device)
            return self._inverse_lambda(logSNR_steps)
        if skip_type == "time_uniform":
            return torch.linspace(tT, t0, N + 1, device=device)
        if skip_type == "time_quadratic":
            t = torch.linspace(t0, tT, 10000000, device=device)
            quadratic_t = torch.sqrt(t)
            quadratic_steps = torch.linspace(quadratic_t[0], quadratic_t[-1], N + 1)
            return torch.flip(
                torch.cat(
                    [
                        t[torch.searchsorted(quadratic_t, quadratic_steps)[:-1]],
                        tT * torch.ones((1,)).to(device),
                    ],
                    dim=0,
                ),
                dims=[0],
            )
        raise ValueError(f"unrecognized skip_type '{skip_type}' occurred")

    def _marginal_log_mean_coef(self, t: Tensor) -> Tensor:
        if self.schedule == "linear":
            return -0.25 * t**2 * (self.beta_1 - self.beta_0) - 0.5 * t * self.beta_0
        if self.schedule == "cosine":
            log_alpha_t = torch.log(
                torch.cos((t + self.cosine_s) / (1.0 + self.cosine_s) * math.pi / 2.0)
            )
            log_alpha_t = log_alpha_t - self.cosine_log_alpha_0
            return log_alpha_t
        raise ValueError(f"unrecognized schedule '{self.schedule}' occurred")

    def _marginal_std(self, t: Tensor) -> Tensor:
        return torch.sqrt(1.0 - torch.exp(2.0 * self._marginal_log_mean_coef(t)))

    def _marginal_lambda(self, t: Tensor) -> Tensor:
        log_mean_coef = self._marginal_log_mean_coef(t)
        log_std = 0.5 * torch.log(1.0 - torch.exp(2.0 * log_mean_coef))
        return log_mean_coef - log_std

    def _inverse_lambda(self, lamb: Tensor) -> Tensor:
        if self.schedule == "linear":
            tmp = (
                2.0
                * (self.beta_1 - self.beta_0)
                * torch.logaddexp(-2.0 * lamb, torch.zeros((1,)).to(lamb))
            )
            delta = self.beta_0**2 + tmp
            return tmp / (torch.sqrt(delta) + self.beta_0) / (self.beta_1 - self.beta_0)
        if self.schedule == "cosine":
            log_alpha = -0.5 * torch.logaddexp(-2.0 * lamb, torch.zeros((1,)).to(lamb))
            t = (
                torch.arccos(torch.exp(log_alpha + self.cosine_log_alpha_0))
                * 2.0
                * (1.0 + self.cosine_s)
                / math.pi
                - self.cosine_s
            )
            return t
        raise ValueError(f"unrecognized schedule '{self.schedule}' occurred")


__all__ = [
    "DPMSolver",
]



import torch

import torch.nn as nn

from torch import Tensor

from ..misc.toolkit import download_model
from ..misc.toolkit import set_requires_grad
from ..models.cv.encoder.backbone.core import Backbone


class ScalingLayer(nn.Module):
    mean: Tensor
    std: Tensor

    def __init__(self) -> None:
        super().__init__()
        mean = torch.tensor([-0.030, -0.088, -0.188])[None, :, None, None]
        std = torch.tensor([0.458, 0.448, 0.450])[None, :, None, None]
        self.register_buffer("mean", mean)
        self.register_buffer("std", std)

    def forward(self, net: Tensor) -> Tensor:
        return (net - self.mean) / self.std


# map input channels to 1
class ChannelMapping(nn.Module):
    def __init__(self, in_channels: int, use_dropout: bool):
        super().__init__()
        blocks = []
        if use_dropout:
            blocks.append(nn.Dropout())
        blocks.append(nn.Conv2d(in_channels, 1, 1, 1, 0, bias=False))
        self.net = nn.Sequential(*blocks)

    def forward(self, net: Tensor) -> Tensor:
        return self.net(net)


def normalize(net: Tensor, eps: float = 1.0e-10) -> Tensor:
    norm = torch.sqrt(torch.sum(net**2, dim=1, keepdim=True))
    return net / (norm + eps)


def spatial_average(net: Tensor, keepdim: bool = True) -> Tensor:
    return net.mean([2, 3], keepdim=keepdim)


class LPIPS(nn.Module):
    def __init__(self, use_dropout: bool = True):
        super().__init__()
        self.scaling_layer = ScalingLayer()
        self.backbone = Backbone("vgg16_full", pretrained=True, requires_grad=False)
        make_mapping = lambda in_nc: ChannelMapping(in_nc, use_dropout)
        self.out_channels = self.backbone.out_channels
        self.mappings = nn.ModuleList(list(map(make_mapping, self.out_channels)))
        self.load_pretrained()
        set_requires_grad(self, False)

    def load_pretrained(self) -> None:
        ckpt = download_model("lpips")
        self.load_state_dict(torch.load(ckpt), strict=False)

    def forward(self, predictions: Tensor, target: Tensor) -> Tensor:
        net0, net1 = map(self.scaling_layer, [predictions, target])
        out0, out1 = map(self.backbone, [net0, net1])
        loss = None
        for i in range(len(self.out_channels)):
            stage = f"stage{i}"
            f0, f1 = out0[stage], out1[stage]
            f0, f1 = map(normalize, [f0, f1])
            diff = (f0 - f1) ** 2
            squeezed = self.mappings[i](diff)
            i_loss = spatial_average(squeezed, keepdim=True)
            if loss is None:
                loss = i_loss
            else:
                loss += i_loss
        return loss


__all__ = [
    "LPIPS",
]



import os
import cflearn

from cflearn.data.ml import MLCarefreeData
from cflearn.api.ml.pipeline import MLPipeline
from cflearn.api.ml.pipeline import MLCarefreePipeline

from ._utils import get_info


if __name__ == "__main__":
    info = get_info()
    kwargs = info.kwargs
    data = info.data
    assert data is not None
    cuda = info.meta["cuda"]
    carefree = isinstance(data, MLCarefreeData)
    m_base = MLCarefreePipeline if carefree else MLPipeline
    m = m_base(**kwargs).fit(data, cuda=cuda)
    m.save(os.path.join(info.workplace, cflearn.ML_PIPELINE_SAVE_NAME))



import torch

from torch import Tensor


def extract_to(array: Tensor, indices: Tensor, num_dim: int) -> Tensor:
    b = indices.shape[0]
    out = array.gather(-1, indices).contiguous()
    return out.view(b, *([1] * (num_dim - 1)))


def get_timesteps(t: int, num: int, device: torch.device) -> Tensor:
    return torch.full((num,), t, device=device, dtype=torch.long)



import os
import json

import numpy as np

from PIL import Image
from tqdm import tqdm
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Callable
from typing import Optional
from cftool.misc import print_warning

from ..basic import CVDataset
from ..basic import Transforms
from ..basic import ImageFolderData
from ..basic import InferenceImageFolderData
from ..basic import ImageFolderDataset
from ..basic import InferenceImageFolderDataset
from ....types import sample_weights_type
from ....constants import INPUT_KEY
from ....constants import STYLE_KEY

try:
    from cfcv.misc.toolkit import to_rgb
except:
    to_rgb = None


class StyleTransferMixin:
    transform: Transforms
    style_paths_file: str = "valid_paths.json"

    def _init_style_paths(self, style_folder: str) -> None:
        self.style_folder = style_folder
        walked = list(os.walk(style_folder))
        extensions = {".jpg", ".png"}
        style_paths_path = os.path.join(style_folder, self.style_paths_file)
        if os.path.isfile(style_paths_path):
            with open(style_paths_path, "r") as f:
                self.style_paths = json.load(f)
            return None
        self.style_paths = []
        for folder, _, files in tqdm(walked, desc="folders", position=0):
            for file in tqdm(files, desc="files", position=1, leave=False):
                if not any(file.endswith(ext) for ext in extensions):
                    continue
                path = os.path.join(folder, file)
                try:
                    to_rgb(Image.open(path)).verify()
                    self.style_paths.append(path)
                except Exception as err:
                    print_warning(f"error occurred ({err}) when reading '{path}'")
                    continue
        with open(style_paths_path, "w") as f:
            json.dump(self.style_paths, f)

    def _inject_style(self, index: int, sample: Dict[str, Any]) -> None:
        style_path = self.style_paths[index % len(self.style_paths)]
        style_img = to_rgb(Image.open(style_path))
        style_arr = np.array(style_img).astype(np.float32) / 255.0
        style_tensor = self.transform({INPUT_KEY: style_arr})[INPUT_KEY]
        sample[STYLE_KEY] = style_tensor


class StyleTransferDataset(ImageFolderDataset, StyleTransferMixin):  # type: ignore
    def __init__(
        self,
        folder: str,
        split: str,
        transform: Optional[Union[str, List[str], Transforms, Callable]],
        transform_config: Optional[Dict[str, Any]] = None,
        extra_label_names: Optional[List[str]] = None,
        lmdb_config: Optional[Dict[str, Any]] = None,
        *,
        style_folder: str,
    ):
        if to_rgb is None:
            raise ValueError("`carefree-cv` is needed for `StyleTransferDataset`")
        super().__init__(
            folder,
            split,
            transform,
            transform_config,
            extra_label_names,
            lmdb_config,
        )
        self._init_style_paths(style_folder)

    def __getitem__(self, index: int) -> Dict[str, Any]:
        sample = super().__getitem__(index)
        self._inject_style(index, sample)
        return sample


class InferenceStyleTransferDataset(InferenceImageFolderDataset, StyleTransferMixin):  # type: ignore
    def __init__(
        self,
        folder: str,
        transform: Optional[Transforms],
        *,
        style_folder: str,
    ):
        if to_rgb is None:
            msg = "`carefree-cv` is needed for `InferenceStyleTransferDataset`"
            raise ValueError(msg)
        super().__init__(folder, transform)
        self._init_style_paths(style_folder)

    def __getitem__(self, index: int) -> Dict[str, Tensor]:
        sample = super().__getitem__(index)
        self._inject_style(index, sample)
        return sample


@ImageFolderData.register("style_transfer")
class StyleTransferData(ImageFolderData):
    def __init__(
        self,
        folder: str,
        style_folder: str,
        *,
        batch_size: int,
        num_workers: int = 0,
        shuffle: bool = True,
        drop_train_last: bool = True,
        transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
        transform_config: Optional[Dict[str, Any]] = None,
        test_shuffle: Optional[bool] = None,
        test_transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
        test_transform_config: Optional[Dict[str, Any]] = None,
        lmdb_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            folder,
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=shuffle,
            drop_train_last=drop_train_last,
            transform=transform,
            transform_config=transform_config,
            test_shuffle=test_shuffle,
            test_transform=test_transform,
            test_transform_config=test_transform_config,
            lmdb_config=lmdb_config,
        )
        self.style_folder = style_folder

    @property
    def info(self) -> Dict[str, Any]:
        d = super().info
        d["style_folder"] = self.style_folder
        return d

    def prepare(self, sample_weights: sample_weights_type) -> None:
        self.train_data = CVDataset(
            StyleTransferDataset(
                self.folder,
                "train",
                self.transform,
                lmdb_config=self.lmdb_config,
                style_folder=self.style_folder,
            )
        )
        self.setup_train_weights()
        self.valid_data = CVDataset(
            StyleTransferDataset(
                self.folder,
                "valid",
                self.test_transform,
                lmdb_config=self.lmdb_config,
                style_folder=self.style_folder,
            )
        )


class InferenceStyleTransferData(InferenceImageFolderData):
    def __init__(
        self,
        folder: str,
        style_folder: str,
        *,
        batch_size: int,
        num_workers: int = 0,
        transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
        transform_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            folder,
            batch_size=batch_size,
            num_workers=num_workers,
            transform=transform,
            transform_config=transform_config,
        )
        self.style_folder = self.kw["style_folder"] = style_folder

    def prepare(self, sample_weights: sample_weights_type) -> None:
        self.dataset = InferenceStyleTransferDataset(
            self.folder,
            self.transform,
            style_folder=self.style_folder,
        )



import torch
import random

import torch.nn as nn

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import List
from typing import Callable
from typing import Optional
from cftool.types import tensor_dict_type

from .discriminators import DiscriminatorBase
from ....protocol import _forward
from ....protocol import ITrainer
from ....protocol import TrainerState
from ....protocol import MetricsOutputs
from ....protocol import WithDeviceMixin
from ....constants import LOSS_KEY
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import PREDICTIONS_KEY
from ...protocols.cv import GaussianGeneratorMixin
from ....misc.toolkit import mode_context
from ....misc.internal_ import CustomTrainStep
from ....misc.internal_ import CustomTrainStepLoss
from ....losses.gan import GANLoss
from ....losses.gan import GANTarget


class GANMixin:
    def _initialize(
        self,
        *,
        num_classes: Optional[int] = None,
        gan_mode: str = "vanilla",
        gan_loss_config: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.num_classes = num_classes
        self.gan_mode = gan_mode
        self.gan_loss = GANLoss(gan_mode)
        if gan_loss_config is None:
            gan_loss_config = {}
        self.lambda_gp = gan_loss_config.get("lambda_gp", 10.0)

    @property
    @abstractmethod
    def g_parameters(self) -> List[nn.Parameter]:
        pass

    @property
    @abstractmethod
    def d_parameters(self) -> List[nn.Parameter]:
        pass

    @abstractmethod
    def _g_losses(
        self,
        batch: tensor_dict_type,
        forward: tensor_dict_type,
    ) -> tensor_dict_type:
        pass

    @abstractmethod
    def _d_losses(
        self,
        batch: tensor_dict_type,
        detached_forward: tensor_dict_type,
    ) -> tensor_dict_type:
        pass

    # utilities

    @property
    def can_reconstruct(self) -> bool:
        return False

    def summary_forward(self, batch_idx: int, batch: tensor_dict_type) -> None:
        self._g_losses(batch, _forward(self, batch_idx, batch, INPUT_KEY))  # type: ignore


class GeneratorStep(CustomTrainStep):
    def loss_fn(
        self,
        m: "OneStageGANMixin",
        trainer: ITrainer,
        batch: tensor_dict_type,
        forward_results: tensor_dict_type,
        **kwargs: Any,
    ) -> CustomTrainStepLoss:
        g_losses = m._g_losses(batch, forward_results)
        g_loss = g_losses.pop(LOSS_KEY)
        g_loss_dict = {k: v.item() for k, v in g_losses.items()}
        g_loss_dict["g"] = g_loss.item()
        return CustomTrainStepLoss(g_loss, g_loss_dict)


class DiscriminatorStep(CustomTrainStep):
    def loss_fn(
        self,
        m: "OneStageGANMixin",
        trainer: ITrainer,
        batch: tensor_dict_type,
        forward_results: tensor_dict_type,
        **kwargs: Any,
    ) -> CustomTrainStepLoss:
        with torch.no_grad():
            detached_forward = {k: v.detach() for k, v in forward_results.items()}
        d_losses = m._d_losses(batch, detached_forward)
        d_loss = d_losses.pop(LOSS_KEY)
        d_loss_dict = {k: v.item() for k, v in d_losses.items()}
        d_loss_dict["g"] = d_loss.item()
        return CustomTrainStepLoss(d_loss, d_loss_dict)


# This mixin should be used with `CustomModule` & `register_custom_module`
class OneStageGANMixin(GANMixin, WithDeviceMixin, metaclass=ABCMeta):
    @property
    def train_steps(self) -> List[CustomTrainStep]:
        return [
            GeneratorStep("core.g_parameters"),
            DiscriminatorStep("core.d_parameters"),
        ]

    def evaluate_step(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: TrainerState,
        weighted_loss_score_fn: Callable[[Dict[str, float]], float],
    ) -> MetricsOutputs:
        forward = _forward(self, batch_idx, batch, INPUT_KEY, state)
        g_losses = self._g_losses(batch, forward)
        # in evaluate step, all tensors are already detached
        d_losses = self._d_losses(batch, forward)
        g_loss = g_losses.pop(LOSS_KEY).item()
        d_loss = d_losses.pop(LOSS_KEY).item()
        loss = g_loss + d_loss
        loss_dict = {"g": g_loss, "d": d_loss, LOSS_KEY: loss}
        loss_dict.update({k: v.item() for k, v in g_losses.items()})
        loss_dict.update({k: v.item() for k, v in d_losses.items()})
        score = weighted_loss_score_fn(loss_dict)
        return MetricsOutputs(score, loss_dict)


class VanillaGANMixin(OneStageGANMixin, GaussianGeneratorMixin, metaclass=ABCMeta):
    def _initialize(  # type: ignore
        self,
        *,
        in_channels: int,
        discriminator: str = "basic",
        discriminator_config: Optional[Dict[str, Any]] = None,
        num_classes: Optional[int] = None,
        gan_mode: str = "vanilla",
        gan_loss_config: Optional[Dict[str, Any]] = None,
    ):
        super()._initialize(
            num_classes=num_classes,
            gan_mode=gan_mode,
            gan_loss_config=gan_loss_config,
        )
        if discriminator_config is None:
            discriminator_config = {}
        discriminator_config["in_channels"] = in_channels
        discriminator_config["num_classes"] = num_classes
        self.discriminator = DiscriminatorBase.make(
            discriminator,
            config=discriminator_config,
        )

    @property
    def d_parameters(self) -> List[nn.Parameter]:
        return list(self.discriminator.parameters())

    def forward(self, batch: tensor_dict_type, **kwargs: Any) -> torch.Tensor:
        z = torch.randn(len(batch[INPUT_KEY]), self.latent_dim, device=self.device)
        return self.decode(z, labels=batch[LABEL_KEY], **kwargs)

    def _g_losses(
        self,
        batch: tensor_dict_type,
        forward: tensor_dict_type,
    ) -> tensor_dict_type:
        labels = batch.get(LABEL_KEY)
        if labels is not None:
            labels = labels.view(-1)
        sampled = forward[PREDICTIONS_KEY]
        pred_fake = self.discriminator(sampled)
        loss_g = self.gan_loss(pred_fake, GANTarget(True, labels))
        return {LOSS_KEY: loss_g}

    def _d_losses(
        self,
        batch: tensor_dict_type,
        detached_forward: tensor_dict_type,
    ) -> tensor_dict_type:
        net = batch[INPUT_KEY]
        labels = batch.get(LABEL_KEY)
        if labels is not None:
            labels = labels.view(-1)
        sampled_tensor = detached_forward[PREDICTIONS_KEY]
        pred_real = self.discriminator(net)
        loss_d_real = self.gan_loss(pred_real, GANTarget(True, labels))
        pred_fake = self.discriminator(sampled_tensor)
        loss_d_fake = self.gan_loss(pred_fake, GANTarget(False, labels))
        d_loss = 0.5 * (loss_d_fake + loss_d_real)
        losses = {"d_fake": loss_d_fake, "d_real": loss_d_real}
        if self.gan_mode == "wgangp":
            eps = random.random()
            merged = eps * net + (1.0 - eps) * sampled_tensor
            with mode_context(self.discriminator, to_train=None, use_grad=True):
                pred_merged = self.discriminator(merged.requires_grad_(True)).output  # type: ignore
                loss_gp = self.gan_loss.loss(merged, pred_merged)
            d_loss = d_loss + self.lambda_gp * loss_gp
            losses["d_gp"] = loss_gp
        losses[LOSS_KEY] = d_loss
        return losses


__all__ = [
    "GANMixin",
    "OneStageGANMixin",
    "VanillaGANMixin",
]
